# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2002-2006 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2002-2004 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import math
from . import unit

#
# classes representing bounding boxes
#

class bbox_pt:
    
    """class for bounding boxes

    This variant requires points in the constructor, and is used for internal
    purposes.

    A bbox for which llx_pt is None represents an empty bbox, i.e., one containing
    no points.
    """

    def __init__(self, llx_pt, lly_pt, urx_pt, ury_pt):
        self.llx_pt = llx_pt
        self.lly_pt = lly_pt
        self.urx_pt = urx_pt
        self.ury_pt = ury_pt

    def __bool__(self):
        return self.llx_pt is not None

    def __add__(self, other):
        """join two bboxes"""
        if self.llx_pt is not None:
            if other.llx_pt is not None:
                return bbox_pt(min(self.llx_pt, other.llx_pt), min(self.lly_pt, other.lly_pt),
                               max(self.urx_pt, other.urx_pt), max(self.ury_pt, other.ury_pt))
            else:
                return bbox_pt(self.llx_pt, self.lly_pt, self.urx_pt, self.ury_pt)
        else:
            return bbox_pt(other.llx_pt, other.lly_pt, other.urx_pt, other.ury_pt)

    def __iadd__(self, other):
        """join two bboxes inplace"""
        if self.llx_pt is not None:
            if other.llx_pt is not None:
                self.llx_pt = min(self.llx_pt, other.llx_pt)
                self.lly_pt = min(self.lly_pt, other.lly_pt)
                self.urx_pt = max(self.urx_pt, other.urx_pt)
                self.ury_pt = max(self.ury_pt, other.ury_pt)
        else:
            self.llx_pt = other.llx_pt
            self.lly_pt = other.lly_pt
            self.urx_pt = other.urx_pt
            self.ury_pt = other.ury_pt
        return self

    def __mul__(self, other):
        """return intersection of two bboxes"""
        if self.llx_pt is not None and other.llx_pt is not None:
            return bbox_pt(max(self.llx_pt, other.llx_pt), max(self.lly_pt, other.lly_pt),
                           min(self.urx_pt, other.urx_pt), min(self.ury_pt, other.ury_pt))
        else:
            return empty()

    def __imul__(self, other):
        """intersect two bboxes in place"""
        if self.llx_pt is not None and other.llx_pt is not None:
            self.llx_pt = max(self.llx_pt, other.llx_pt)
            self.lly_pt = max(self.lly_pt, other.lly_pt)
            self.urx_pt = min(self.urx_pt, other.urx_pt)
            self.ury_pt = min(self.ury_pt, other.ury_pt)
        elif other.llx_pt is None:
            self.llx_pt = None
        return self

    def copy(self):
        return bbox_pt(self.llx_pt, self.lly_pt, self.urx_pt, self.ury_pt)

    def set(self, other):
        self.llx_pt = other.llx_pt
        self.lly_pt = other.lly_pt
        self.urx_pt = other.urx_pt
        self.ury_pt = other.ury_pt

    def lowrestuple_pt(self):
        if self.llx_pt is None:
            raise ValueError("Cannot return low-res tuple for empty bbox")
        return (math.floor(self.llx_pt), math.floor(self.lly_pt),
                math.ceil(self.urx_pt), math.ceil(self.ury_pt))

    def highrestuple_pt(self):
        if self.llx_pt is None:
            raise ValueError("Cannot return high-res tuple for empty bbox")
        return (self.llx_pt, self.lly_pt, self.urx_pt, self.ury_pt)

    def intersects(self, other):
        """check, if two bboxes intersect eachother"""
        if self.llx_pt is None or other.llx_pt is None:
            return 0
        else:
            return not (self.llx_pt > other.urx_pt or
                        self.lly_pt > other.ury_pt or
                        self.urx_pt < other.llx_pt or
                        self.ury_pt < other.lly_pt)

    def includepoint_pt(self, x_pt, y_pt):
        if self.llx_pt is None:
            self.llx_pt = self.urx_pt = x_pt
            self.lly_pt = self.ury_pt = y_pt
        else:
            self.llx_pt = min(self.llx_pt, x_pt)
            self.lly_pt = min(self.lly_pt, y_pt)
            self.urx_pt = max(self.urx_pt, x_pt)
            self.ury_pt = max(self.ury_pt, y_pt)

    def transform(self, trafo):
        """transform bbox in place by trafo"""
        if self.llx_pt is None:
            return
        # we have to transform all four corner points of the bbox
        llx_pt, lly_pt = trafo.apply_pt(self.llx_pt, self.lly_pt)
        lrx_pt, lry_pt = trafo.apply_pt(self.urx_pt, self.lly_pt)
        urx_pt, ury_pt = trafo.apply_pt(self.urx_pt, self.ury_pt)
        ulx_pt, uly_pt = trafo.apply_pt(self.llx_pt, self.ury_pt)

        # Now, by sorting, we obtain the lower left and upper right corner
        # of the new bounding box.
        self.llx_pt = min(llx_pt, lrx_pt, urx_pt, ulx_pt)
        self.lly_pt = min(lly_pt, lry_pt, ury_pt, uly_pt)
        self.urx_pt = max(llx_pt, lrx_pt, urx_pt, ulx_pt)
        self.ury_pt = max(lly_pt, lry_pt, ury_pt, uly_pt)

    def transformed(self, trafo):
        """return bbox transformed by trafo"""
        if self.llx_pt is None:
            return empty()
        # we have to transform all four corner points of the bbox
        llx_pt, lly_pt = trafo.apply_pt(self.llx_pt, self.lly_pt)
        lrx_pt, lry_pt = trafo.apply_pt(self.urx_pt, self.lly_pt)
        urx_pt, ury_pt = trafo.apply_pt(self.urx_pt, self.ury_pt)
        ulx_pt, uly_pt = trafo.apply_pt(self.llx_pt, self.ury_pt)

        # Now, by sorting, we obtain the lower left and upper right corner
        # of the new bounding box.
        return bbox_pt(min(llx_pt, lrx_pt, urx_pt, ulx_pt), min(lly_pt, lry_pt, ury_pt, uly_pt),
                       max(llx_pt, lrx_pt, urx_pt, ulx_pt), max(lly_pt, lry_pt, ury_pt, uly_pt))

    def enlarge_pt(self, all_pt=0, bottom_pt=None, left_pt=None, top_pt=None, right_pt=None):
        """enlarge bbox in place by the given amounts in pts

        all is used, if bottom, left, top and/or right are not given.

        """
        if self.llx_pt is None:
            return
        if bottom_pt is None:
           bottom_pt = all_pt
        if left_pt is None:
           left_pt = all_pt
        if top_pt is None:
           top_pt = all_pt
        if right_pt is None:
           right_pt = all_pt
        self.llx_pt -= left_pt
        self.lly_pt -= bottom_pt
        self.urx_pt += right_pt
        self.ury_pt += top_pt

    def enlarged_pt(self, all_pt=0, bottom_pt=None, left_pt=None, top_pt=None, right_pt=None):
        """return bbox enlarged by the given amounts in pts

        all is used, if bottom, left, top and/or right are not given.

        """
        if self.llx_pt is None:
            return empty()
        if bottom_pt is None:
           bottom_pt = all_pt
        if left_pt is None:
           left_pt = all_pt
        if top_pt is None:
           top_pt = all_pt
        if right_pt is None:
           right_pt = all_pt
        return bbox_pt(self.llx_pt-left_pt, self.lly_pt-bottom_pt, self.urx_pt+right_pt, self.ury_pt+top_pt)

    def enlarge(self, all=0, bottom=None, left=None, top=None, right=None):
        """enlarge bbox in place

        all is used, if bottom, left, top and/or right are not given.

        """
        if self.llx_pt is None:
            return
        bottom_pt = left_pt = top_pt = right_pt = unit.topt(all)
        if bottom is not None:
           bottom_pt = unit.topt(bottom)
        if left is not None:
           left_pt = unit.topt(left)
        if top is not None:
           top_pt = unit.topt(top)
        if right is not None:
           right_pt = unit.topt(right)
        self.llx_pt -= left_pt
        self.lly_pt -= bottom_pt
        self.urx_pt += right_pt
        self.ury_pt += top_pt

    def enlarged(self, all=0, bottom=None, left=None, top=None, right=None):
        """return bbox enlarged

        all is used, if bottom, left, top and/or right are not given.

        """
        if self.llx_pt is None:
            return empty()
        bottom_pt = left_pt = top_pt = right_pt = unit.topt(all)
        if bottom is not None:
           bottom_pt = unit.topt(bottom)
        if left is not None:
           left_pt = unit.topt(left)
        if top is not None:
           top_pt = unit.topt(top)
        if right is not None:
           right_pt = unit.topt(right)
        return bbox_pt(self.llx_pt-left_pt, self.lly_pt-bottom_pt, self.urx_pt+right_pt, self.ury_pt+top_pt)

    def rect(self):
        """return rectangle corresponding to bbox"""
        from . import path
        if self.llx_pt is None:
            raise ValueError("Cannot return path for empty bbox")
        return path.rect_pt(self.llx_pt, self.lly_pt, self.urx_pt-self.llx_pt, self.ury_pt-self.lly_pt)

    path = rect

    def height_pt(self):
        """return height of bbox in pts"""
        if self.llx_pt is None:
            raise ValueError("Cannot return heigth of empty bbox")
        return self.ury_pt-self.lly_pt

    def width_pt(self):
        """return width of bbox in pts"""
        if self.llx_pt is None:
            raise ValueError("Cannot return width of empty bbox")
        return self.urx_pt-self.llx_pt

    def top_pt(self):
        """return top coordinate of bbox in pts"""
        if self.llx_pt is None:
            raise ValueError("Cannot return top coordinate of empty bbox")
        return self.ury_pt

    def bottom_pt(self):
        """return bottom coordinate of bbox in pts"""
        if self.llx_pt is None:
            raise ValueError("Cannot return bottom coordinate of empty bbox")
        return self.lly_pt

    def left_pt(self):
        """return left coordinate of bbox in pts"""
        if self.llx_pt is None:
            raise ValueError("Cannot return left coordinate of empty bbox")
        return self.llx_pt

    def right_pt(self):
        """return right coordinate of bbox in pts"""
        if self.llx_pt is None:
            raise ValueError("Cannot return right coordinate of empty bbox")
        return self.urx_pt

    def center_pt(self):
        """return coordinates of the center of the bbox in pts"""
        if self.llx_pt is None:
            raise ValueError("Cannot return center coordinates of empty bbox")
        return 0.5 * (self.llx_pt+self.urx_pt), 0.5 * (self.lly_pt+self.ury_pt)

    def height(self):
        """return height of bbox"""
        return self.height_pt() * unit.t_pt

    def width(self):
        """return width of bbox"""
        return self.width_pt() * unit.t_pt

    def top(self):
        """return top coordinate of bbox"""
        return self.ury_pt * unit.t_pt

    def bottom(self):
        """return bottom coordinate of bbox"""
        return self.lly_pt * unit.t_pt

    def left(self):
        """return left coordinate of bbox"""
        return self.llx_pt * unit.t_pt

    def right(self):
        """return right coordinate of bbox"""
        return self.urx_pt * unit.t_pt

    def center(self):
        """return coordinates of the center of the bbox"""
        centerx_pt, centery_pt = self.center_pt()
        return centerx_pt * unit.t_pt, centery_pt * unit.t_pt


class bbox(bbox_pt):

    """class for bounding boxes"""

    def __init__(self, llx_pt, lly_pt, urx_pt, ury_pt):
        llx_pt = unit.topt(llx_pt)
        lly_pt = unit.topt(lly_pt)
        urx_pt = unit.topt(urx_pt)
        ury_pt = unit.topt(ury_pt)
        bbox_pt.__init__(self, llx_pt, lly_pt, urx_pt, ury_pt)


class empty(bbox_pt):

    """empty bounding box, i.e., one containing no point"""
    def __init__(self):
        bbox_pt.__init__(self, None, None, None, None)



from pyx import *

c = canvas.canvas()

rect1 = path.path(path.moveto(0, 0), path.lineto(1, 0),
                  path.moveto(1, 0), path.lineto(1, 1),
                  path.moveto(1, 1), path.lineto(0, 1),
                  path.moveto(0, 1), path.lineto(0, 0))
rect2 = path.path(path.moveto(2, 0), path.lineto(3, 0),
                  path.lineto(3, 1), path.lineto(2, 1),
                  path.lineto(2, 0))
rect3 = path.path(path.moveto(4, 0), path.lineto(5, 0),
                  path.lineto(5, 1), path.lineto(4, 1),
                  path.closepath())

c.stroke(rect1, [style.linewidth.THICK])
c.stroke(rect2, [style.linewidth.THICK])
c.stroke(rect3, [style.linewidth.THICK])

c.writeEPSfile("pathitem")
c.writePDFfile("pathitem")
c.writeSVGfile("pathitem")



# suggested by Chris Spencer

from pyx import *

g = graph.graphxy(width=8)
# either provide lists of the individual coordinates
g.plot(graph.data.values(x=list(range(10)), y=list(range(10))))
# or provide one list containing the whole points
g.plot(graph.data.points(list(zip(range(10), range(10))), x=1, y=2))
g.writeEPSfile("points")
g.writePDFfile("points")
g.writeSVGfile("points")



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2005-2011 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2005-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import logging, sys
from . import bbox, pswriter, pdfwriter, svgwriter, trafo, style, unit

logger = logging.getLogger("pyx")

class paperformat:

    def __init__(self, width, height, name=None):
        self.width = width
        self.height = height
        self.name = name

paperformat.A5 = paperformat(148.5 * unit.t_mm, 210  * unit.t_mm, "A5")
paperformat.A4 = paperformat(210 * unit.t_mm, 297  * unit.t_mm, "A4")
paperformat.A3 = paperformat(297 * unit.t_mm, 420  * unit.t_mm, "A3")
paperformat.A2 = paperformat(420 * unit.t_mm, 594  * unit.t_mm, "A2")
paperformat.A1 = paperformat(594 * unit.t_mm, 840  * unit.t_mm, "A1")
paperformat.A0 = paperformat(840 * unit.t_mm, 1188 * unit.t_mm, "A0")
paperformat.A0b = paperformat(910 * unit.t_mm, 1370 * unit.t_mm, None) # dedicated to our friends in Augsburg
paperformat.Letter = paperformat(8.5 * unit.t_inch, 11 * unit.t_inch, "Letter")
paperformat.Legal = paperformat(8.5 * unit.t_inch, 14 * unit.t_inch, "Legal")

def _paperformatfromstring(name):
    return getattr(paperformat, name.capitalize())


class page:

    def __init__(self, canvas, pagename=None, paperformat=None, rotated=0, centered=1, fittosize=0,
                 margin=1*unit.t_cm, bboxenlarge=1*unit.t_pt, bbox=None):
        self.canvas = canvas
        self.pagename = pagename
        # support for deprecated string specification of paper formats
        try:
            paperformat + ""
        except Exception:
            self.paperformat = paperformat
        else:
            self.paperformat = _paperformatfromstring(paperformat)
            logger.warning("specification of paperformat by string is deprecated, use document.paperformat.%s instead" % paperformat.capitalize())

        self.rotated = rotated
        self.centered = centered
        self.fittosize = fittosize
        self.margin = margin
        self.bboxenlarge = bboxenlarge
        self.pagebbox = bbox

    def _process(self, processMethod, contentfile, writer, context, registry, bbox):
        # usually, it is the bbox of the canvas enlarged by self.bboxenlarge, but
        # it might be a different bbox as specified in the page constructor
        assert not bbox
        if self.pagebbox:
            bbox.set(self.pagebbox)
        else:
            bbox.set(self.canvas.bbox()) # this bbox is not accurate
            bbox.enlarge(self.bboxenlarge)

        # check whether we expect a page trafo and use a temporary canvas to insert the
        # page canvas
        if self.paperformat and (self.rotated or self.centered or self.fittosize) and bbox:
            # calculate the pagetrafo
            paperwidth, paperheight = self.paperformat.width, self.paperformat.height

            # center (optionally rotated) output on page
            if self.rotated:
                pagetrafo = trafo.rotate(90).translated(paperwidth, 0)
                if self.centered or self.fittosize:
                    if not self.fittosize and (bbox.height() > paperwidth or bbox.width() > paperheight):
                        logger.warning("content exceeds the papersize")
                    pagetrafo = pagetrafo.translated(-0.5*(paperwidth - bbox.height()) + bbox.bottom(),
                                                      0.5*(paperheight - bbox.width()) - bbox.left())
            else:
                if not self.fittosize and (bbox.width() > paperwidth or bbox.height() > paperheight):
                    logger.warning("content exceeds the papersize")
                pagetrafo = trafo.translate(0.5*(paperwidth - bbox.width())  - bbox.left(),
                                            0.5*(paperheight - bbox.height()) - bbox.bottom())

            if self.fittosize:

                if 2*self.margin > paperwidth or 2*self.margin > paperheight:
                    raise ValueError("Margins too broad for selected paperformat. Aborting.")

                paperwidth -= 2 * self.margin
                paperheight -= 2 * self.margin

                # scale output to pagesize - margins
                if self.rotated:
                    sfactor = min(unit.topt(paperheight)/bbox.width_pt(), unit.topt(paperwidth)/bbox.height_pt())
                else:
                    sfactor = min(unit.topt(paperwidth)/bbox.width_pt(), unit.topt(paperheight)/bbox.height_pt())

                pagetrafo = pagetrafo.scaled(sfactor, sfactor, self.margin + 0.5*paperwidth, self.margin + 0.5*paperheight)

            bbox.transform(pagetrafo)
            from . import canvas as canvasmodule
            cc = canvasmodule.canvas()
            cc.insert(self.canvas, [pagetrafo])
        else:
            cc = self.canvas

        if processMethod != "processSVG":
            # for SVG we write the pyx defaults as part of the svg node attributes in the writer
            getattr(style.linewidth.normal, processMethod)(contentfile, writer, context, registry)
        if self.pagebbox:
            bbox = bbox.copy() # don't alter the bbox provided to the constructor -> use a copy
        getattr(cc, processMethod)(contentfile, writer, context, registry, bbox)

    def processPS(self, *args):
        self._process("processPS", *args)

    def processPDF(self, *args):
        self._process("processPDF", *args)

    def processSVG(self, *args):
        self._process("processSVG", *args)


class _noclose:

    def __init__(self, f):
        self.f = f

    def __enter__(self):
        return self.f

    def __exit__(self, type, value, tb):
        pass


def _outputstream(file, suffix):
    if file is None:
        if not sys.argv[0].endswith(".py"):
            raise RuntimeError("could not auto-guess filename")
        return open("%s.%s" % (sys.argv[0][:-3], suffix), "wb")
    if file == "-":
        return _noclose(sys.stdout.buffer)
    try:
        file.write(b"")
    except Exception:
        if not file.endswith(".%s" % suffix):
            return open("%s.%s" % (file, suffix), "wb")
        return open(file, "wb")
    else:
        return _noclose(file)


class document:

    """holds a collection of page instances which are output as pages of a document"""

    def __init__(self, pages=None):
        if pages is None:
            self.pages = []
        else:
            self.pages = pages

    def append(self, page):
        self.pages.append(page)

    def writeEPSfile(self, file=None, **kwargs):
        with _outputstream(file, "eps") as f:
            pswriter.EPSwriter(self, f, **kwargs)

    def writePSfile(self, file=None, **kwargs):
        with _outputstream(file, "ps") as f:
            pswriter.PSwriter(self, f, **kwargs)

    def writePDFfile(self, file=None, **kwargs):
        with _outputstream(file, "pdf") as f:
            pdfwriter.PDFwriter(self, f, **kwargs)

    def writeSVGfile(self, file=None, **kwargs):
        with _outputstream(file, "svg") as f:
            svgwriter.SVGwriter(self, f, **kwargs)

    def writetofile(self, filename, **kwargs):
        for suffix, method in [("eps", pswriter.EPSwriter),
                               ("ps", pswriter.PSwriter),
                               ("pdf", pdfwriter.PDFwriter),
                               ("svg", svgwriter.SVGwriter)]:
            if filename.endswith(".{}".format(suffix)):
                with open(filename, "wb") as f:
                    method(self, f, **kwargs)
                return
        raise ValueError("unknown file extension")




import math
from pyx import *

c = canvas.canvas()

g1 = c.insert(graph.graphxy(width=8))
g1.plot(graph.data.function("y(x)=2*exp(-30*x)-exp(-3*x)", min=0, max=1))

g2 = c.insert(graph.graphxy(width=8, ypos=g1.height+0.5,
                            x=graph.axis.linkedaxis(g1.axes["x"])))
g2.plot(graph.data.function("y(x)=cos(20*x)*exp(-2*x)"))

c.writeEPSfile("link")
c.writePDFfile("link")
c.writeSVGfile("link")



from pyx import *

d = graph.data.points([(1,  0.3),
                       (2, -0.7),
                       (3, -0.3),
                       (4,  0.8),
                       (5,  0.5)], x=1, y=2)

g = graph.graphxy(width=8)
g.plot(d, [graph.style.histogram()])
g.writeEPSfile("histogram")
g.writePDFfile("histogram")
g.writeSVGfile("histogram")



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2007-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import struct, re
from . import metric


ansiglyphs = {"space": 32,
              "exclam": 33,
              "quotedbl": 34,
              "numbersign": 35,
              "dollar": 36,
              "percent": 37,
              "ampersand": 38,
              "quotesingle": 39,
              "parenleft": 40,
              "parenright": 41,
              "asterisk": 42,
              "plus": 43,
              "comma": 44,
              "hyphen": 45,
              "period": 46,
              "slash": 47,
              "zero": 48,
              "one": 49,
              "two": 50,
              "three": 51,
              "four": 52,
              "five": 53,
              "six": 54,
              "seven": 55,
              "eight": 56,
              "nine": 57,
              "colon": 58,
              "semicolon": 59,
              "less": 60,
              "equal": 61,
              "greater": 62,
              "question": 63,
              "at": 64,
              "A": 65,
              "B": 66,
              "C": 67,
              "D": 68,
              "E": 69,
              "F": 70,
              "G": 71,
              "H": 72,
              "I": 73,
              "J": 74,
              "K": 75,
              "L": 76,
              "M": 77,
              "N": 78,
              "O": 79,
              "P": 80,
              "Q": 81,
              "R": 82,
              "S": 83,
              "T": 84,
              "U": 85,
              "V": 86,
              "W": 87,
              "X": 88,
              "Y": 89,
              "Z": 90,
              "bracketleft": 91,
              "backslash": 92,
              "bracketright": 93,
              "asciicircum": 94,
              "underscore": 95,
              "grave": 96,
              "a": 97,
              "b": 98,
              "c": 99,
              "d": 100,
              "e":101,
              "f":102,
              "g":103,
              "h":104,
              "i":105,
              "j":106,
              "k":107,
              "l":108,
              "m":109,
              "n":110,
              "o":111,
              "p":112,
              "q":113,
              "r":114,
              "s":115,
              "t":116,
              "u":117,
              "v":118,
              "w":119,
              "x":120,
              "y":121,
              "z":122,
              "braceleft":123,
              "bar":124,
              "braceright":125,
              "asciitilde":126,
              "bullet":127,
              "Euro":128,
              "bullet":129,
              "quotesinglbase":130,
              "florin":131,
              "quotedblbase":132,
              "ellipsis":133,
              "dagger":134,
              "daggerdbl":135,
              "circumflex":136,
              "perthousand":137,
              "Scaron":138,
              "guilsinglleft":139,
              "OE":140,
              "bullet":141,
              "Zcaron":142,
              "bullet":143,
              "bullet":144,
              "quoteleft":145,
              "quoteright":146,
              "quotedblleft":147,
              "quotedblright":148,
              "bullet":149,
              "endash":150,
              "emdash":151,
              "tilde":152,
              "trademark":153,
              "scaron":154,
              "guilsinglright":155,
              "oe":156,
              "bullet":157,
              "zcaron":158,
              "Ydieresis":159,
              "space":160,
              "exclamdown":161,
              "cent":162,
              "sterling":163,
              "currency":164,
              "yen":165,
              "brokenbar":166,
              "section":167,
              "dieresis":168,
              "copyright":169,
              "ordfeminine":170,
              "guillemotleft":171,
              "logicalnot":172,
              "hyphen":173,
              "registered":174,
              "macron":175,
              "degree":176,
              "plusminus":177,
              "twosuperior":178,
              "threesuperior":179,
              "acute":180,
              "mu":181,
              "paragraph":182,
              "periodcentered":183,
              "cedilla":184,
              "onesuperior":185,
              "ordmasculine":186,
              "guillemotright":187,
              "onequarter":188,
              "onehalf":189,
              "threequarters":190,
              "questiondown":191,
              "Agrave":192,
              "Aacute":193,
              "Acircumflex":194,
              "Atilde":195,
              "Adieresis":196,
              "Aring":197,
              "AE":198,
              "Ccedilla":199,
              "Egrave":200,
              "Eacute":201,
              "Ecircumflex":202,
              "Edieresis":203,
              "Igrave":204,
              "Iacute":205,
              "Icircumflex":206,
              "Idieresis":207,
              "Eth":208,
              "Ntilde":209,
              "Ograve":210,
              "Oacute":211,
              "Ocircumflex":212,
              "Otilde":213,
              "Odieresis":214,
              "multiply":215,
              "Oslash":216,
              "Ugrave":217,
              "Uacute":218,
              "Ucircumflex":219,
              "Udieresis":220,
              "Yacute":221,
              "Thorn":222,
              "germandbls":223,
              "agrave":224,
              "aacute":225,
              "acircumflex":226,
              "atilde":227,
              "adieresis":228,
              "aring":229,
              "ae":230,
              "ccedilla":231,
              "egrave":232,
              "eacute":233,
              "ecircumflex":234,
              "edieresis":235,
              "igrave":236,
              "iacute":237,
              "icircumflex":238,
              "idieresis":239,
              "eth":240,
              "ntilde":241,
              "ograve":242,
              "oacute":243,
              "ocircumflex":244,
              "otilde":245,
              "odieresis":246,
              "divide":247,
              "oslash":248,
              "ugrave":249,
              "uacute":250,
              "ucircumflex":251,
              "udieresis":252,
              "yacute":253,
              "thorn":254,
              "ydieresis":255}


fontbboxpattern = re.compile(r"/FontBBox\s*\{\s*(?P<fontbbox>(-?[0-9.]+)\s+(-?[0-9.]+)\s+(-?[0-9.]+)\s+(-?[0-9.]+))\s*\}\s*(readonly\s+)?def")


def _readNullString(file):
    s = []
    c = file.read(1)
    while c and c != b"\0":
        s.append(c)
        c = file.read(1)
    return b"".join(s)


class PFMfile(metric.metric):

    def __init__(self, file, t1file):
        # pfm is rather incomplete, the t1file instance can be used to fill the gap
        (self.dfVersion, self.dfSize, self.dfCopyright, self.dfType,
         self.dfPoint, self.dfVertRes, self.dfHorizRes, self.dfAscent,
         self.dfInternalLeading, self.dfExternalLeading, self.dfItalic,
         self.dfUnderline, self.dfStrikeOut, self.dfWeight,
         self.dfCharSet, self.dfPixWidth, self.dfPixHeight,
         self.dfPitchAndFamily, self.dfAvgWidth, self.dfMaxWidth,
         self.dfFirstChar, self.dfLastChar, self.dfDefaultChar,
         self.dfBreakChar, self.dfWidthBytes, self.dfDevice, self.dfFace,
         self.dfBitsPointer, self.dfBitsOffset) = struct.unpack("<HL60s7H3BHB2HB2H4BH4L", file.read(117))
        self.dfCopyright = self.dfCopyright.split(b"\x00", 1)[0]
        (self.dfSizeFields, self.dfExtMetricsOffset, self.dfExtentTable,
         self.dfOriginTable, self.dfPairKernTable, self.dfTrackKernTable,
         self.dfDriverInfo, self.dfReserved) = struct.unpack("<H7L", file.read(30))
        if self.dfDevice == 0:
            raise ValueError("DeviceName is required for Type1 pfm files.")
        file.seek(self.dfDevice)
        self.deviceName = _readNullString(file)
        if self.deviceName.lower() != b"postscript":
            raise ValueError("Can process pfm files for PostScript fonts only.")
        if self.dfVersion != 0x100:
            raise ValueError("Invalid pfm file version.")
        if self.dfType != 0x81:
            raise ValueError("Not a Type1 pfm file.")
        if self.dfFace == 0:
            raise ValueError("FaceName is required for Type1 pfm files.")
        if self.dfExtMetricsOffset == 0:
            raise ValueError("ExtTextMetrics is required for Type1 pfm files.")
        if self.dfExtentTable == 0:
            raise ValueError("ExtentTable is required for Type1 pfm files.")
        if self.dfOriginTable != 0:
            raise ValueError("OriginTable is forbidden for Type1 pfm files.")
        if self.dfDriverInfo == 0:
            raise ValueError("DriverInfo is required for Type1 pfm files.")
        # assert self.dfReserved == 0 (must be zero according to the spec, but we don't care)
        file.seek(self.dfExtMetricsOffset)
        (etmSize, self.etmPointSize, self.etmOrientation,
         self.etmMasterHeight, self.etmMinScale, self.etmMaxScale,
         self.etmMasterUnits, self.etmCapHeight, self.etmXHeight,
         self.etmLowerCaseAscent, self.etmLowerCaseDescent, self.etmSlant,
         self.etmSuperScript, self.etmSubScript, self.etmSuperScriptSize,
         self.etmSubScriptSize, self.etmUnderlineOffset,
         self.etmUnderlineWidth, self.etmDoubleUpperUnderlineOffset,
         self.etmDoubleLowerUnderlineOffset, self.etmDoubleUpperUnderlineWidth,
         self.etmDoubleLowerUnderlineWidth, self.etmStrikeOutOffset,
         self.etmStrikeOutWidth, self.etmKernPairs, self.etmKernTracks) = struct.unpack("<24h2H", file.read(52))
        file.seek(self.dfFace)
        self.faceName = _readNullString(file)
        file.seek(self.dfDriverInfo)
        self.driverInfo = _readNullString(file)
        file.seek(self.dfExtentTable)
        count = self.dfLastChar - self.dfFirstChar + 1
        self.widths = struct.unpack("<%dH" % count, file.read(2*count))
        self.kernpairs = []
        self.kernpairsdict = {}
        if self.dfPairKernTable:
            file.seek(self.dfPairKernTable)
            pairs, = struct.unpack("<H", file.read(2))
            if pairs != self.etmKernPairs:
                raise ValueError("number of kerning pairs mismatch in pfm file.")
            for i in range(self.etmKernPairs):
                kpFirst, kpSecond, kpKernAmount = struct.unpack("<BBh", file.read(4))
                self.kernpairs.append((kpFirst, kpSecond, kpKernAmount))
                self.kernpairsdict[(kpFirst, kpSecond)] = kpKernAmount
        self.trackkerns = []
        if self.dfTrackKernTable:
            file.seek(self.dfTrackKernTable)
            items, = struct.unpack("<H", file.read(2))
            if items != self.etmKernTracks:
                raise ValueError("number of kerning tracks mismatch in pfm file.")
            for i in range(self.etmKernTracks):
                # each item consists of the tuple ktDegree, ktMinSize, ktMinAmount, ktMaxSize, ktMaxAmount
                self.trackkerns.append(struct.unpack("<hhhhh", file.read(10)))
        if self.dfCharSet:
            if not t1file.encoding:
                t1file._encoding()
            self.glyphs = dict([(glyph, i) for i, glyph in enumerate(t1file.encoding) if glyph != None])
        else:
            self.glyphs = ansiglyphs
        self.t1file = t1file

    def width_ds(self, glyphname):
        return self.widths[self.glyphs[glyphname]-self.dfFirstChar]

    def width_pt(self, glyphnames, size_pt):
        return sum([self.widths[self.glyphs[glyphname]-self.dfFirstChar] for glyphname in glyphnames])*size_pt/1000.0

    def height_pt(self, glyphnames, size_pt):
        return self.dfAscent*size_pt/1000.0

    def depth_pt(self, glyphnames, size_pt):
        return -self.etmLowerCaseDescent*size_pt/1000.0

    def resolvekernings(self, glyphnames, size_pt=None):
        result = [None]*(2*len(glyphnames)-1)
        for i, glyphname in enumerate(glyphnames):
            result[2*i] = glyphname
        return result

    def resolvekernings(self, glyphnames, size_pt=None):
        result = [None]*(2*len(glyphnames)-1)
        for i, glyphname in enumerate(glyphnames):
            result[2*i] = glyphname
            if i:
                amount = self.kernpairsdict.get((self.glyphs[glyphnames[i-1]], self.glyphs[glyphname]))
                if amount:
                    if size_pt is not None:
                        result[2*i-1] = amount*size_pt/1000.0
                    else:
                        result[2*i-1] = amount
        return result

    def writePDFfontinfo(self, file, seriffont=False, symbolfont=True):
        flags = 0
        if self.dfMaxWidth == self.dfAvgWidth:
            flags += 1<<0
        if seriffont:
            flags += 1<<1
        if symbolfont:
            flags += 1<<2
        else:
            flags += 1<<5
        if self.dfItalic:
            flags += 1<<6
        file.write("/Flags %d\n" % flags)
        file.write("/ItalicAngle %d\n" % (self.etmSlant/10))
        file.write("/Ascent %d\n" % self.dfAscent)
        file.write("/Descent %d\n" % -self.etmLowerCaseDescent)
        file.write("/FontBBox [%s]\n" % fontbboxpattern.search(self.t1file.data1).groups('fontbbox')[0])
        file.write("/CapHeight %d\n" % self.etmCapHeight)
        if self.dfWeight >= 600:
            stemv = 120
        else:
            stemv = 70
        file.write("/StemV %d\n" % stemv)




from pyx import *

c = canvas.canvas()

# apply global TeX setting
text.preamble(r"\parindent=0pt")
w = 1.2 # an appropriate parbox width for spam & eggs

# vertical alignments by margins
c.stroke(path.line(0, 4, 8, 4), [style.linewidth.THin])
c.text(0, 4, r"spam \& eggs", [text.parbox(w), text.valign.top])
c.text(3, 4, r"spam \& eggs", [text.parbox(w), text.valign.middle])
c.text(6, 4, r"spam \& eggs", [text.parbox(w), text.valign.bottom])

# vertical alignments by baselines
c.stroke(path.line(0, 2, 8, 2), [style.linewidth.THin])
c.text(0, 2, r"spam \& eggs", [text.parbox(w, baseline=text.parbox.top)])
c.text(3, 2, r"spam \& eggs", [text.parbox(w, baseline=text.parbox.middle)])
c.text(6, 2, r"spam \& eggs", [text.parbox(w, baseline=text.parbox.bottom)])

# vertical shifts
c.stroke(path.line(0, 0, 8, 0), [style.linewidth.THin])
c.text(0, 0, r"x=0", [text.mathmode, text.vshift.topzero])
c.text(2, 0, r"x=0", [text.mathmode, text.vshift.middlezero])
c.text(4, 0, r"x=0", [text.mathmode, text.vshift.bottomzero])
c.text(6, 0, r"x=0", [text.mathmode, text.vshift.mathaxis])

c.writeEPSfile("valign")
c.writePDFfile("valign")
c.writeSVGfile("valign")



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2006-2011 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2007-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import string
from . import metric

unicodestring = {" ": "space",
                 "!": "exclam",
                 "\"": "quotedbl",
                 "#": "numbersign",
                 "$": "dollar",
                 "%": "percent",
                 "&": "ampersand",
                 "'": "quotesingle",
                 "(": "parenleft",
                 ")": "parenright",
                 "*": "asterisk",
                 "+": "plus",
                 ",": "comma",
                 "-": "hyphen",
                 ".": "period",
                 "/": "slash",
                 "0": "zero",
                 "1": "one",
                 "2": "two",
                 "3": "three",
                 "4": "four",
                 "5": "five",
                 "6": "six",
                 "7": "seven",
                 "8": "eight",
                 "9": "nine",
                 ":": "colon",
                 ";": "semicolon",
                 "<": "less",
                 "=": "equal",
                 ">": "greater",
                 "?": "question",
                 "@": "at",
                 "A": "A",
                 "B": "B",
                 "C": "C",
                 "D": "D",
                 "E": "E",
                 "F": "F",
                 "G": "G",
                 "H": "H",
                 "I": "I",
                 "J": "J",
                 "K": "K",
                 "L": "L",
                 "M": "M",
                 "N": "N",
                 "O": "O",
                 "P": "P",
                 "Q": "Q",
                 "R": "R",
                 "S": "S",
                 "T": "T",
                 "U": "U",
                 "V": "V",
                 "W": "W",
                 "X": "X",
                 "Y": "Y",
                 "Z": "Z",
                 "[": "bracketleft",
                 "\\": "backslash",
                 "]": "bracketright",
                 "^": "asciicircum",
                 "_": "underscore",
                 "`": "grave",
                 "a": "a",
                 "b": "b",
                 "c": "c",
                 "d": "d",
                 "e": "e",
                 "f": "f",
                 "g": "g",
                 "h": "h",
                 "i": "i",
                 "j": "j",
                 "k": "k",
                 "l": "l",
                 "m": "m",
                 "n": "n",
                 "o": "o",
                 "p": "p",
                 "q": "q",
                 "r": "r",
                 "s": "s",
                 "t": "t",
                 "u": "u",
                 "v": "v",
                 "w": "w",
                 "x": "x",
                 "y": "y",
                 "z": "z",
                 "{": "braceleft",
                 "|": "bar",
                 "}": "braceright",
                 "~": "asciitilde",
                 "\xa0": "space",
                 "\xa1": "exclamdown",
                 "\xa2": "cent",
                 "\xa3": "sterling",
                 "\xa4": "currency",
                 "\xa5": "yen",
                 "\xa6": "brokenbar",
                 "\xa7": "section",
                 "\xa8": "dieresis",
                 "\xa9": "copyright",
                 "\xaa": "ordfeminine",
                 "\xab": "guillemotleft",
                 "\xac": "logicalnot",
                 "\xad": "hyphen",
                 "\xae": "registered",
                 "\xaf": "macron",
                 "\xb0": "degree",
                 "\xb1": "plusminus",
                 "\xb4": "acute",
                 "\xb6": "paragraph",
                 "\xb7": "periodcentered",
                 "\xb8": "cedilla",
                 "\xba": "ordmasculine",
                 "\xbb": "guillemotright",
                 "\xbc": "onequarter",
                 "\xbd": "onehalf",
                 "\xbe": "threequarters",
                 "\xbf": "questiondown",
                 "\xc0": "Agrave",
                 "\xc1": "Aacute",
                 "\xc2": "Acircumflex",
                 "\xc3": "Atilde",
                 "\xc4": "Adieresis",
                 "\xc5": "Aring",
                 "\xc6": "AE",
                 "\xc7": "Ccedilla",
                 "\xc8": "Egrave",
                 "\xc9": "Eacute",
                 "\xca": "Ecircumflex",
                 "\xcb": "Edieresis",
                 "\xcc": "Igrave",
                 "\xcd": "Iacute",
                 "\xce": "Icircumflex",
                 "\xcf": "Idieresis",
                 "\xd0": "Eth",
                 "\xd1": "Ntilde",
                 "\xd2": "Ograve",
                 "\xd3": "Oacute",
                 "\xd4": "Ocircumflex",
                 "\xd5": "Otilde",
                 "\xd6": "Odieresis",
                 "\xd7": "multiply",
                 "\xd8": "Oslash",
                 "\xd9": "Ugrave",
                 "\xda": "Uacute",
                 "\xdb": "Ucircumflex",
                 "\xdc": "Udieresis",
                 "\xdd": "Yacute",
                 "\xde": "Thorn",
                 "\xdf": "germandbls",
                 "\xe0": "agrave",
                 "\xe1": "aacute",
                 "\xe2": "acircumflex",
                 "\xe3": "atilde",
                 "\xe4": "adieresis",
                 "\xe5": "aring",
                 "\xe6": "ae",
                 "\xe7": "ccedilla",
                 "\xe8": "egrave",
                 "\xe9": "eacute",
                 "\xea": "ecircumflex",
                 "\xeb": "edieresis",
                 "\xec": "igrave",
                 "\xed": "iacute",
                 "\xee": "icircumflex",
                 "\xef": "idieresis",
                 "\xf0": "eth",
                 "\xf1": "ntilde",
                 "\xf2": "ograve",
                 "\xf3": "oacute",
                 "\xf4": "ocircumflex",
                 "\xf5": "otilde",
                 "\xf6": "odieresis",
                 "\xf7": "divide",
                 "\xf8": "oslash",
                 "\xf9": "ugrave",
                 "\xfa": "uacute",
                 "\xfb": "ucircumflex",
                 "\xfc": "udieresis",
                 "\xfd": "yacute",
                 "\xfe": "thorn",
                 "\xff": "ydieresis",
                 "\u0100": "Amacron",
                 "\u0101": "amacron",
                 "\u0102": "Abreve",
                 "\u0103": "abreve",
                 "\u0104": "Aogonek",
                 "\u0105": "aogonek",
                 "\u0106": "Cacute",
                 "\u0107": "cacute",
                 "\u0108": "Ccircumflex",
                 "\u0109": "ccircumflex",
                 "\u010a": "Cdotaccent",
                 "\u010b": "cdotaccent",
                 "\u010c": "Ccaron",
                 "\u010d": "ccaron",
                 "\u010e": "Dcaron",
                 "\u010f": "dcaron",
                 "\u0110": "Dcroat",
                 "\u0111": "dcroat",
                 "\u0112": "Emacron",
                 "\u0113": "emacron",
                 "\u0114": "Ebreve",
                 "\u0115": "ebreve",
                 "\u0116": "Edotaccent",
                 "\u0117": "edotaccent",
                 "\u0118": "Eogonek",
                 "\u0119": "eogonek",
                 "\u011a": "Ecaron",
                 "\u011b": "ecaron",
                 "\u011c": "Gcircumflex",
                 "\u011d": "gcircumflex",
                 "\u011e": "Gbreve",
                 "\u011f": "gbreve",
                 "\u0120": "Gdotaccent",
                 "\u0121": "gdotaccent",
                 "\u0122": "Gcommaaccent",
                 "\u0123": "gcommaaccent",
                 "\u0124": "Hcircumflex",
                 "\u0125": "hcircumflex",
                 "\u0126": "Hbar",
                 "\u0127": "hbar",
                 "\u0128": "Itilde",
                 "\u0129": "itilde",
                 "\u012a": "Imacron",
                 "\u012b": "imacron",
                 "\u012c": "Ibreve",
                 "\u012d": "ibreve",
                 "\u012e": "Iogonek",
                 "\u012f": "iogonek",
                 "\u0130": "Idotaccent",
                 "\u0131": "dotlessi",
                 "\u0132": "IJ",
                 "\u0133": "ij",
                 "\u0134": "Jcircumflex",
                 "\u0135": "jcircumflex",
                 "\u0136": "Kcommaaccent",
                 "\u0137": "kcommaaccent",
                 "\u0138": "kgreenlandic",
                 "\u0139": "Lacute",
                 "\u013a": "lacute",
                 "\u013b": "Lcommaaccent",
                 "\u013c": "lcommaaccent",
                 "\u013d": "Lcaron",
                 "\u013e": "lcaron",
                 "\u013f": "Ldot",
                 "\u0140": "ldot",
                 "\u0141": "Lslash",
                 "\u0142": "lslash",
                 "\u0143": "Nacute",
                 "\u0144": "nacute",
                 "\u0145": "Ncommaaccent",
                 "\u0146": "ncommaaccent",
                 "\u0147": "Ncaron",
                 "\u0148": "ncaron",
                 "\u0149": "napostrophe",
                 "\u014a": "Eng",
                 "\u014b": "eng",
                 "\u014c": "Omacron",
                 "\u014d": "omacron",
                 "\u014e": "Obreve",
                 "\u014f": "obreve",
                 "\u0150": "Ohungarumlaut",
                 "\u0151": "ohungarumlaut",
                 "\u0152": "OE",
                 "\u0153": "oe",
                 "\u0154": "Racute",
                 "\u0155": "racute",
                 "\u0156": "Rcommaaccent",
                 "\u0157": "rcommaaccent",
                 "\u0158": "Rcaron",
                 "\u0159": "rcaron",
                 "\u015a": "Sacute",
                 "\u015b": "sacute",
                 "\u015c": "Scircumflex",
                 "\u015d": "scircumflex",
                 "\u015e": "Scedilla",
                 "\u015f": "scedilla",
                 "\u0160": "Scaron",
                 "\u0161": "scaron",
                 "\u0162": "Tcommaaccent",
                 "\u0163": "tcommaaccent",
                 "\u0164": "Tcaron",
                 "\u0165": "tcaron",
                 "\u0166": "Tbar",
                 "\u0167": "tbar",
                 "\u0168": "Utilde",
                 "\u0169": "utilde",
                 "\u016a": "Umacron",
                 "\u016b": "umacron",
                 "\u016c": "Ubreve",
                 "\u016d": "ubreve",
                 "\u016e": "Uring",
                 "\u016f": "uring",
                 "\u0170": "Uhungarumlaut",
                 "\u0171": "uhungarumlaut",
                 "\u0172": "Uogonek",
                 "\u0173": "uogonek",
                 "\u0174": "Wcircumflex",
                 "\u0175": "wcircumflex",
                 "\u0176": "Ycircumflex",
                 "\u0177": "ycircumflex",
                 "\u0178": "Ydieresis",
                 "\u0179": "Zacute",
                 "\u017a": "zacute",
                 "\u017b": "Zdotaccent",
                 "\u017c": "zdotaccent",
                 "\u017d": "Zcaron",
                 "\u017e": "zcaron",
                 "\u017f": "longs",
                 "\u0192": "florin",
                 "\u01a0": "Ohorn",
                 "\u01a1": "ohorn",
                 "\u01af": "Uhorn",
                 "\u01b0": "uhorn",
                 "\u01e6": "Gcaron",
                 "\u01e7": "gcaron",
                 "\u01fa": "Aringacute",
                 "\u01fb": "aringacute",
                 "\u01fc": "AEacute",
                 "\u01fd": "aeacute",
                 "\u01fe": "Oslashacute",
                 "\u01ff": "oslashacute",
                 "\u0218": "Scommaaccent",
                 "\u0219": "scommaaccent",
                 "\u02bc": "afii57929",
                 "\u02bd": "afii64937",
                 "\u02c6": "circumflex",
                 "\u02c7": "caron",
                 "\u02c9": "macron",
                 "\u02d8": "breve",
                 "\u02d9": "dotaccent",
                 "\u02da": "ring",
                 "\u02db": "ogonek",
                 "\u02dc": "tilde",
                 "\u02dd": "hungarumlaut",
                 "\u0300": "gravecomb",
                 "\u0301": "acutecomb",
                 "\u0303": "tildecomb",
                 "\u0309": "hookabovecomb",
                 "\u0323": "dotbelowcomb",
                 "\u0384": "tonos",
                 "\u0385": "dieresistonos",
                 "\u0386": "Alphatonos",
                 "\u0387": "anoteleia",
                 "\u0388": "Epsilontonos",
                 "\u0389": "Etatonos",
                 "\u038a": "Iotatonos",
                 "\u038c": "Omicrontonos",
                 "\u038e": "Upsilontonos",
                 "\u038f": "Omegatonos",
                 "\u0390": "iotadieresistonos",
                 "\u0391": "Alpha",
                 "\u0392": "Beta",
                 "\u0393": "Gamma",
                 "\u0394": "Delta",
                 "\u0395": "Epsilon",
                 "\u0396": "Zeta",
                 "\u0397": "Eta",
                 "\u0398": "Theta",
                 "\u0399": "Iota",
                 "\u039a": "Kappa",
                 "\u039b": "Lambda",
                 "\u039c": "Mu",
                 "\u039d": "Nu",
                 "\u039e": "Xi",
                 "\u039f": "Omicron",
                 "\u03a0": "Pi",
                 "\u03a1": "Rho",
                 "\u03a3": "Sigma",
                 "\u03a4": "Tau",
                 "\u03a5": "Upsilon",
                 "\u03a6": "Phi",
                 "\u03a7": "Chi",
                 "\u03a8": "Psi",
                 "\u03a9": "Omega",
                 "\u03aa": "Iotadieresis",
                 "\u03ab": "Upsilondieresis",
                 "\u03ac": "alphatonos",
                 "\u03ad": "epsilontonos",
                 "\u03ae": "etatonos",
                 "\u03af": "iotatonos",
                 "\u03b0": "upsilondieresistonos",
                 "\u03b1": "alpha",
                 "\u03b2": "beta",
                 "\u03b3": "gamma",
                 "\u03b4": "delta",
                 "\u03b5": "epsilon",
                 "\u03b6": "zeta",
                 "\u03b7": "eta",
                 "\u03b8": "theta",
                 "\u03b9": "iota",
                 "\u03ba": "kappa",
                 "\u03bb": "lambda",
                 "\u03bc": "mu",
                 "\u03bd": "nu",
                 "\u03be": "xi",
                 "\u03bf": "omicron",
                 "\u03c0": "pi",
                 "\u03c1": "rho",
                 "\u03c2": "sigma1",
                 "\u03c3": "sigma",
                 "\u03c4": "tau",
                 "\u03c5": "upsilon",
                 "\u03c6": "phi",
                 "\u03c7": "chi",
                 "\u03c8": "psi",
                 "\u03c9": "omega",
                 "\u03ca": "iotadieresis",
                 "\u03cb": "upsilondieresis",
                 "\u03cc": "omicrontonos",
                 "\u03cd": "upsilontonos",
                 "\u03ce": "omegatonos",
                 "\u03d1": "theta1",
                 "\u03d2": "Upsilon1",
                 "\u03d5": "phi1",
                 "\u03d6": "omega1",
                 "\u0401": "afii10023",
                 "\u0402": "afii10051",
                 "\u0403": "afii10052",
                 "\u0404": "afii10053",
                 "\u0405": "afii10054",
                 "\u0406": "afii10055",
                 "\u0407": "afii10056",
                 "\u0408": "afii10057",
                 "\u0409": "afii10058",
                 "\u040a": "afii10059",
                 "\u040b": "afii10060",
                 "\u040c": "afii10061",
                 "\u040e": "afii10062",
                 "\u040f": "afii10145",
                 "\u0410": "afii10017",
                 "\u0411": "afii10018",
                 "\u0412": "afii10019",
                 "\u0413": "afii10020",
                 "\u0414": "afii10021",
                 "\u0415": "afii10022",
                 "\u0416": "afii10024",
                 "\u0417": "afii10025",
                 "\u0418": "afii10026",
                 "\u0419": "afii10027",
                 "\u041a": "afii10028",
                 "\u041b": "afii10029",
                 "\u041c": "afii10030",
                 "\u041d": "afii10031",
                 "\u041e": "afii10032",
                 "\u041f": "afii10033",
                 "\u0420": "afii10034",
                 "\u0421": "afii10035",
                 "\u0422": "afii10036",
                 "\u0423": "afii10037",
                 "\u0424": "afii10038",
                 "\u0425": "afii10039",
                 "\u0426": "afii10040",
                 "\u0427": "afii10041",
                 "\u0428": "afii10042",
                 "\u0429": "afii10043",
                 "\u042a": "afii10044",
                 "\u042b": "afii10045",
                 "\u042c": "afii10046",
                 "\u042d": "afii10047",
                 "\u042e": "afii10048",
                 "\u042f": "afii10049",
                 "\u0430": "afii10065",
                 "\u0431": "afii10066",
                 "\u0432": "afii10067",
                 "\u0433": "afii10068",
                 "\u0434": "afii10069",
                 "\u0435": "afii10070",
                 "\u0436": "afii10072",
                 "\u0437": "afii10073",
                 "\u0438": "afii10074",
                 "\u0439": "afii10075",
                 "\u043a": "afii10076",
                 "\u043b": "afii10077",
                 "\u043c": "afii10078",
                 "\u043d": "afii10079",
                 "\u043e": "afii10080",
                 "\u043f": "afii10081",
                 "\u0440": "afii10082",
                 "\u0441": "afii10083",
                 "\u0442": "afii10084",
                 "\u0443": "afii10085",
                 "\u0444": "afii10086",
                 "\u0445": "afii10087",
                 "\u0446": "afii10088",
                 "\u0447": "afii10089",
                 "\u0448": "afii10090",
                 "\u0449": "afii10091",
                 "\u044a": "afii10092",
                 "\u044b": "afii10093",
                 "\u044c": "afii10094",
                 "\u044d": "afii10095",
                 "\u044e": "afii10096",
                 "\u044f": "afii10097",
                 "\u0451": "afii10071",
                 "\u0452": "afii10099",
                 "\u0453": "afii10100",
                 "\u0454": "afii10101",
                 "\u0455": "afii10102",
                 "\u0456": "afii10103",
                 "\u0457": "afii10104",
                 "\u0458": "afii10105",
                 "\u0459": "afii10106",
                 "\u045a": "afii10107",
                 "\u045b": "afii10108",
                 "\u045c": "afii10109",
                 "\u045e": "afii10110",
                 "\u045f": "afii10193",
                 "\u0462": "afii10146",
                 "\u0463": "afii10194",
                 "\u0472": "afii10147",
                 "\u0473": "afii10195",
                 "\u0474": "afii10148",
                 "\u0475": "afii10196",
                 "\u0490": "afii10050",
                 "\u0491": "afii10098",
                 "\u04d9": "afii10846",
                 "\u05b0": "afii57799",
                 "\u05b1": "afii57801",
                 "\u05b2": "afii57800",
                 "\u05b3": "afii57802",
                 "\u05b4": "afii57793",
                 "\u05b5": "afii57794",
                 "\u05b6": "afii57795",
                 "\u05b7": "afii57798",
                 "\u05b8": "afii57797",
                 "\u05b9": "afii57806",
                 "\u05bb": "afii57796",
                 "\u05bc": "afii57807",
                 "\u05bd": "afii57839",
                 "\u05be": "afii57645",
                 "\u05bf": "afii57841",
                 "\u05c0": "afii57842",
                 "\u05c1": "afii57804",
                 "\u05c2": "afii57803",
                 "\u05c3": "afii57658",
                 "\u05d0": "afii57664",
                 "\u05d1": "afii57665",
                 "\u05d2": "afii57666",
                 "\u05d3": "afii57667",
                 "\u05d4": "afii57668",
                 "\u05d5": "afii57669",
                 "\u05d6": "afii57670",
                 "\u05d7": "afii57671",
                 "\u05d8": "afii57672",
                 "\u05d9": "afii57673",
                 "\u05da": "afii57674",
                 "\u05db": "afii57675",
                 "\u05dc": "afii57676",
                 "\u05dd": "afii57677",
                 "\u05de": "afii57678",
                 "\u05df": "afii57679",
                 "\u05e0": "afii57680",
                 "\u05e1": "afii57681",
                 "\u05e2": "afii57682",
                 "\u05e3": "afii57683",
                 "\u05e4": "afii57684",
                 "\u05e5": "afii57685",
                 "\u05e6": "afii57686",
                 "\u05e7": "afii57687",
                 "\u05e8": "afii57688",
                 "\u05e9": "afii57689",
                 "\u05ea": "afii57690",
                 "\u05f0": "afii57716",
                 "\u05f1": "afii57717",
                 "\u05f2": "afii57718",
                 "\u060c": "afii57388",
                 "\u061b": "afii57403",
                 "\u061f": "afii57407",
                 "\u0621": "afii57409",
                 "\u0622": "afii57410",
                 "\u0623": "afii57411",
                 "\u0624": "afii57412",
                 "\u0625": "afii57413",
                 "\u0626": "afii57414",
                 "\u0627": "afii57415",
                 "\u0628": "afii57416",
                 "\u0629": "afii57417",
                 "\u062a": "afii57418",
                 "\u062b": "afii57419",
                 "\u062c": "afii57420",
                 "\u062d": "afii57421",
                 "\u062e": "afii57422",
                 "\u062f": "afii57423",
                 "\u0630": "afii57424",
                 "\u0631": "afii57425",
                 "\u0632": "afii57426",
                 "\u0633": "afii57427",
                 "\u0634": "afii57428",
                 "\u0635": "afii57429",
                 "\u0636": "afii57430",
                 "\u0637": "afii57431",
                 "\u0638": "afii57432",
                 "\u0639": "afii57433",
                 "\u063a": "afii57434",
                 "\u0640": "afii57440",
                 "\u0641": "afii57441",
                 "\u0642": "afii57442",
                 "\u0643": "afii57443",
                 "\u0644": "afii57444",
                 "\u0645": "afii57445",
                 "\u0646": "afii57446",
                 "\u0647": "afii57470",
                 "\u0648": "afii57448",
                 "\u0649": "afii57449",
                 "\u064a": "afii57450",
                 "\u064b": "afii57451",
                 "\u064c": "afii57452",
                 "\u064d": "afii57453",
                 "\u064e": "afii57454",
                 "\u064f": "afii57455",
                 "\u0650": "afii57456",
                 "\u0651": "afii57457",
                 "\u0652": "afii57458",
                 "\u0660": "afii57392",
                 "\u0661": "afii57393",
                 "\u0662": "afii57394",
                 "\u0663": "afii57395",
                 "\u0664": "afii57396",
                 "\u0665": "afii57397",
                 "\u0666": "afii57398",
                 "\u0667": "afii57399",
                 "\u0668": "afii57400",
                 "\u0669": "afii57401",
                 "\u066a": "afii57381",
                 "\u066d": "afii63167",
                 "\u0679": "afii57511",
                 "\u067e": "afii57506",
                 "\u0686": "afii57507",
                 "\u0688": "afii57512",
                 "\u0691": "afii57513",
                 "\u0698": "afii57508",
                 "\u06a4": "afii57505",
                 "\u06af": "afii57509",
                 "\u06ba": "afii57514",
                 "\u06d2": "afii57519",
                 "\u06d5": "afii57534",
                 "\u1e80": "Wgrave",
                 "\u1e81": "wgrave",
                 "\u1e82": "Wacute",
                 "\u1e83": "wacute",
                 "\u1e84": "Wdieresis",
                 "\u1e85": "wdieresis",
                 "\u1ef2": "Ygrave",
                 "\u1ef3": "ygrave",
                 "\u200c": "afii61664",
                 "\u200d": "afii301",
                 "\u200e": "afii299",
                 "\u200f": "afii300",
                 "\u2012": "figuredash",
                 "\u2013": "endash",
                 "\u2014": "emdash",
                 "\u2015": "afii00208",
                 "\u2017": "underscoredbl",
                 "\u2018": "quoteleft",
                 "\u2019": "quoteright",
                 "\u201a": "quotesinglbase",
                 "\u201b": "quotereversed",
                 "\u201c": "quotedblleft",
                 "\u201d": "quotedblright",
                 "\u201e": "quotedblbase",
                 "\u2020": "dagger",
                 "\u2021": "daggerdbl",
                 "\u2022": "bullet",
                 "\u2024": "onedotenleader",
                 "\u2025": "twodotenleader",
                 "\u2026": "ellipsis",
                 "\u202c": "afii61573",
                 "\u202d": "afii61574",
                 "\u202e": "afii61575",
                 "\u2030": "perthousand",
                 "\u2032": "minute",
                 "\u2033": "second",
                 "\u2039": "guilsinglleft",
                 "\u203a": "guilsinglright",
                 "\u203c": "exclamdbl",
                 "\u2044": "fraction",
                 "\u20a1": "colonmonetary",
                 "\u20a3": "franc",
                 "\u20a4": "lira",
                 "\u20a7": "peseta",
                 "\u20aa": "afii57636",
                 "\u20ab": "dong",
                 "\u20ac": "Euro",
                 "\u2105": "afii61248",
                 "\u2111": "Ifraktur",
                 "\u2113": "afii61289",
                 "\u2116": "afii61352",
                 "\u2118": "weierstrass",
                 "\u211c": "Rfraktur",
                 "\u211e": "prescription",
                 "\u2122": "trademark",
                 "\u212e": "estimated",
                 "\u2135": "aleph",
                 "\u2153": "onethird",
                 "\u2154": "twothirds",
                 "\u215b": "oneeighth",
                 "\u215c": "threeeighths",
                 "\u215d": "fiveeighths",
                 "\u215e": "seveneighths",
                 "\u2190": "arrowleft",
                 "\u2191": "arrowup",
                 "\u2192": "arrowright",
                 "\u2193": "arrowdown",
                 "\u2194": "arrowboth",
                 "\u2195": "arrowupdn",
                 "\u21a8": "arrowupdnbse",
                 "\u21b5": "carriagereturn",
                 "\u21d0": "arrowdblleft",
                 "\u21d1": "arrowdblup",
                 "\u21d2": "arrowdblright",
                 "\u21d3": "arrowdbldown",
                 "\u21d4": "arrowdblboth",
                 "\u2200": "universal",
                 "\u2202": "partialdiff",
                 "\u2203": "existential",
                 "\u2205": "emptyset",
                 "\u2207": "gradient",
                 "\u2208": "element",
                 "\u2209": "notelement",
                 "\u220b": "suchthat",
                 "\u220f": "product",
                 "\u2211": "summation",
                 "\u2212": "minus",
                 "\u2215": "fraction",
                 "\u2217": "asteriskmath",
                 "\u2219": "periodcentered",
                 "\u221a": "radical",
                 "\u221d": "proportional",
                 "\u221e": "infinity",
                 "\u221f": "orthogonal",
                 "\u2220": "angle",
                 "\u2227": "logicaland",
                 "\u2228": "logicalor",
                 "\u2229": "intersection",
                 "\u222a": "union",
                 "\u222b": "integral",
                 "\u2234": "therefore",
                 "\u223c": "similar",
                 "\u2245": "congruent",
                 "\u2248": "approxequal",
                 "\u2260": "notequal",
                 "\u2261": "equivalence",
                 "\u2264": "lessequal",
                 "\u2265": "greaterequal",
                 "\u2282": "propersubset",
                 "\u2283": "propersuperset",
                 "\u2284": "notsubset",
                 "\u2286": "reflexsubset",
                 "\u2287": "reflexsuperset",
                 "\u2295": "circleplus",
                 "\u2297": "circlemultiply",
                 "\u22a5": "perpendicular",
                 "\u22c5": "dotmath",
                 "\u2302": "house",
                 "\u2310": "revlogicalnot",
                 "\u2320": "integraltp",
                 "\u2321": "integralbt",
                 "\u2329": "angleleft",
                 "\u232a": "angleright",
                 "\u2500": "SF100000",
                 "\u2502": "SF110000",
                 "\u250c": "SF010000",
                 "\u2510": "SF030000",
                 "\u2514": "SF020000",
                 "\u2518": "SF040000",
                 "\u251c": "SF080000",
                 "\u2524": "SF090000",
                 "\u252c": "SF060000",
                 "\u2534": "SF070000",
                 "\u253c": "SF050000",
                 "\u2550": "SF430000",
                 "\u2551": "SF240000",
                 "\u2552": "SF510000",
                 "\u2553": "SF520000",
                 "\u2554": "SF390000",
                 "\u2555": "SF220000",
                 "\u2556": "SF210000",
                 "\u2557": "SF250000",
                 "\u2558": "SF500000",
                 "\u2559": "SF490000",
                 "\u255a": "SF380000",
                 "\u255b": "SF280000",
                 "\u255c": "SF270000",
                 "\u255d": "SF260000",
                 "\u255e": "SF360000",
                 "\u255f": "SF370000",
                 "\u2560": "SF420000",
                 "\u2561": "SF190000",
                 "\u2562": "SF200000",
                 "\u2563": "SF230000",
                 "\u2564": "SF470000",
                 "\u2565": "SF480000",
                 "\u2566": "SF410000",
                 "\u2567": "SF450000",
                 "\u2568": "SF460000",
                 "\u2569": "SF400000",
                 "\u256a": "SF540000",
                 "\u256b": "SF530000",
                 "\u256c": "SF440000",
                 "\u2580": "upblock",
                 "\u2584": "dnblock",
                 "\u2588": "block",
                 "\u258c": "lfblock",
                 "\u2590": "rtblock",
                 "\u2591": "ltshade",
                 "\u2592": "shade",
                 "\u2593": "dkshade",
                 "\u25a0": "filledbox",
                 "\u25a1": "H22073",
                 "\u25aa": "H18543",
                 "\u25ab": "H18551",
                 "\u25ac": "filledrect",
                 "\u25b2": "triagup",
                 "\u25ba": "triagrt",
                 "\u25bc": "triagdn",
                 "\u25c4": "triaglf",
                 "\u25ca": "lozenge",
                 "\u25cb": "circle",
                 "\u25cf": "H18533",
                 "\u25d8": "invbullet",
                 "\u25d9": "invcircle",
                 "\u25e6": "openbullet",
                 "\u263a": "smileface",
                 "\u263b": "invsmileface",
                 "\u263c": "sun",
                 "\u2640": "female",
                 "\u2642": "male",
                 "\u2660": "spade",
                 "\u2663": "club",
                 "\u2665": "heart",
                 "\u2666": "diamond",
                 "\u266a": "musicalnote",
                 "\u266b": "musicalnotedbl",
                 "\ufb01": "fi",
                 "\ufb02": "fl"}

class AFMError(Exception):
    pass

# reader states
_READ_START       = 0
_READ_MAIN        = 1
_READ_DIRECTION   = 2
_READ_CHARMETRICS = 3
_READ_KERNDATA    = 4
_READ_TRACKKERN   = 5
_READ_KERNPAIRS   = 6
_READ_COMPOSITES  = 7
_READ_END         = 8

# various parsing functions
def _parseint(s):
    try:
        return int(s)
    except Exception:
        raise AFMError("Expecting int, got '%s'" % s)

def _parsehex(s):
    try:
        if s[0] != "<" or s[-1] != ">":
            raise AFMError()
        return int(s[1:-1], 16)
    except Exception:
        raise AFMError("Expecting hexadecimal int, got '%s'" % s)

def _parsefloat(s):
    try:
        # Against the AFM spec, some AFM files contain floats with "," instead of "." as decimal separators (e.g. DejaVuSans.afm)
        # Gracefuly accept such input.
        s = s.replace(",", ".")
        return float(s)
    except Exception:
        raise AFMError("Expecting float, got '%s'" % s)

def _parsefloats(s, nos):
    try:
        # Gracefuly accept input with "," instead of ".", similarly to _parsefloat()
        s = s.replace(",", ".")
        numbers = s.split()
        result = list(map(float, numbers))
        if len(result) != nos:
            raise AFMError()
    except Exception:
        raise AFMError("Expecting list of %d numbers, got '%s'" % (s, nos))
    return result

def _parsestr(s):
    # XXX: check for invalid characters in s
    return s

def _parsebool(s):
    s = s.rstrip()
    if s == "true":
       return True
    elif s == "false":
       return False
    else:
        raise AFMError("Expecting boolean, got '%s'" % s)


class AFMcharmetrics:
    def __init__(self, code, widths=None, vvector=None, name=None, bbox=None, ligatures=None):
        self.code = code
        if widths is None:
            self.widths = [None] * 2
        else:
            self.widths = widths
        self.vvector = vvector
        self.name = name
        self.bbox = bbox
        if ligatures is None:
            self.ligatures = []
        else:
            self.ligatures = ligatures


class AFMtrackkern:
    def __init__(self, degree, min_ptsize, min_kern, max_ptsize, max_kern):
        self.degree = degree
        self.min_ptsize = min_ptsize
        self.min_kern = min_kern
        self.max_ptsize = max_ptsize
        self.max_kern = max_kern


class AFMkernpair:
    def __init__(self, name1, name2, x, y):
        self.name1 = name1
        self.name2 = name2
        self.x = x
        self.y = y


class AFMcomposite:
    def __init__(self, name, parts):
        self.name = name
        self.parts = parts


class AFMfile(metric.metric):

    def __init__(self, file):
       self.afmversion = None                   # version, required
       self.metricssets = 0                     # int, optional
       self.fontname = None                     # str, required
       self.fullname = None                     # str, optional
       self.familyname = None                   # str, optional
       self.weight = None                       # str, optional
       self.fontbbox = None                     # 4 floats, required
       self.version = None                      # str, optional
       self.notice = None                       # str, optional
       self.encodingscheme = None               # str, optional
       self.mappingscheme = None                # int, optional (not present in base font programs)
       self.escchar = None                      # int, required if mappingscheme == 3
       self.characterset = None                 # str, optional
       self.characters = None                   # int, optional
       self.isbasefont = True                   # bool, optional
       self.vvector = None                      # 2 floats, required if metricssets == 2
       self.isfixedv = None                     # bool, default: true if vvector present, false otherwise
       self.capheight = None                    # float, optional
       self.xheight = None                      # float, optional
       self.ascender = None                     # float, optional
       self.descender = None                    # float, optional
       self.stdhw = None                        # float, optional
       self.stdvw = None                        # float, optional
       self.underlinepositions = [None] * 2     # int, optional (for each direction)
       self.underlinethicknesses = [None] * 2   # float, optional (for each direction)
       self.italicangles = [None] * 2           # float, optional (for each direction)
       self.charwidths = [None] * 2             # 2 floats, optional (for each direction)
       self.isfixedpitchs = [None] * 2          # bool, optional (for each direction)
       self.expected_entries = None             # if set, internal variable to verify number of expected entries in a section
       self.charmetrics = None                  # list of character metrics information, optional
       self.charmetricsdict = {}                # helper dictionary mapping glyph names to character metrics information
       self.trackkerns = None                   # list of track kernings, optional
       self.kernpairs = [None] * 2              # list of list of kerning pairs (for each direction), optional
       self.kernpairsdict = {}                  # helper dictionary mapping glyph names to kerning pairs, first direction
       self.kernpairsdict1 = {}                 # helper dictionary mapping glyph names to kerning pairs, second direction
       self.composites = None                   # list of composite character data sets, optional
       self.parse(file)
       if self.isfixedv is None:
           self.isfixedv = self.vvector is not None
       # XXX we should check the constraints on some parameters

    # the following methods process a line when the reader is in the corresponding
    # state and return the new state
    def _processline_start(self, line):
        key, args = line.split(None, 1)
        if key != "StartFontMetrics":
            raise AFMError("Expecting StartFontMetrics, no found")
        self.afmversion = tuple(map(int, args.split(".")))
        return _READ_MAIN, None

    def _processline_main(self, line):
        try:
            key, args = line.split(None, 1)
        except ValueError:
            key = line
            args = None
        if key == "Comment":
            return _READ_MAIN, None
        elif key == "MetricsSets":
            self.metricssets = _parseint(args)
            if direction is not None:
                raise AFMError("MetricsSets not allowed after first (implicit) StartDirection")
        elif key == "FontName":
            self.fontname = _parsestr(args)
        elif key == "FullName":
            self.fullname = _parsestr(args)
        elif key == "FamilyName":
            self.familyname = _parsestr(args)
        elif key == "Weight":
            self.weight = _parsestr(args)
        elif key == "FontBBox":
            self.fontbbox = _parsefloats(args, 4)
        elif key == "Version":
            if args is not None:
                self.version = _parsestr(args)
        elif key == "Notice":
            self.notice = _parsestr(args)
        elif key == "EncodingScheme":
            self.encodingscheme = _parsestr(args)
        elif key == "MappingScheme":
            self.mappingscheme = _parseint(args)
        elif key == "EscChar":
            self.escchar = _parseint(args)
        elif key == "CharacterSet":
            self.characterset = _parsestr(args)
        elif key == "Characters":
            self.characters = _parseint(args)
        elif key == "IsBaseFont":
            self.isbasefont = _parsebool(args)
        elif key == "VVector":
            self.vvector = _parsefloats(args, 2)
        elif key == "IsFixedV":
            self.isfixedv = _parsebool(args)
        elif key == "CapHeight":
            self.capheight = _parsefloat(args)
        elif key == "XHeight":
            self.xheight = _parsefloat(args)
        elif key == "Ascender":
            self.ascender = _parsefloat(args)
        elif key == "Descender":
            self.descender = _parsefloat(args)
        elif key == "StdHW":
            self.stdhw = _parsefloat(args)
        elif key == "StdVW":
            self.stdvw = _parsefloat(args)
        elif key == "StartDirection":
            direction = _parseint(args)
            if 0 <= direction <= 2:
                return _READ_DIRECTION, direction
            else:
                raise AFMError("Wrong direction number %d" % direction)
        elif (key == "UnderlinePosition" or key == "UnderlineThickness" or key == "ItalicAngle" or
              key == "Charwidth" or key == "IsFixedPitch"):
            # we implicitly entered a direction section, so we should process the line again
            return self._processline_direction(line, 0)
        elif key == "StartCharMetrics":
            if self.charmetrics is not None:
                raise AFMError("Multiple character metrics sections")
            if self.afmversion >= (2, 0):
                self.expected_entries = _parseint(args)
            else:
                self.expected_entries = None
            self.charmetrics = []
            return _READ_CHARMETRICS, 0
        elif key == "StartKernData":
            return _READ_KERNDATA, None
        elif key == "StartComposites":
            if self.composites is not None:
                raise AFMError("Multiple composite character data sections")
            if args is not None:
                self.expected_entries = _parseint(args)
            else:
                self.expected_entries = None
            self.composites = []
            return _READ_COMPOSITES, 0
        elif key == "EndFontMetrics":
            return _READ_END, None
        elif key[0] in string.ascii_lowercase:
            # ignoring private commands
            pass
        else:
            # and according to the AFM specs also all other unknown keys
            pass
        return _READ_MAIN, None

    def _processline_direction(self, line, direction):
        try:
            key, args = line.split(None, 1)
        except ValueError:
            key = line.strip()
        if key == "UnderlinePosition":
            self.underlinepositions[direction] = _parsefloat(args)
        elif key == "UnderlineThickness":
            self.underlinethicknesses[direction] = _parsefloat(args)
        elif key == "ItalicAngle":
            self.italicangles[direction] = _parsefloat(args)
        elif key == "Charwidth":
            self.charwidths[direction] = _parsefloats(args, 2)
        elif key == "IsFixedPitch":
            self.isfixedpitchs[direction] = _parsebool(args)
        elif key == "EndDirection":
            return _READ_MAIN, None
        else:
            # we assume that we are implicitly leaving the direction section again,
            # so try to reprocess the line in the header reader state
            return self._processline_main(line)
        return _READ_DIRECTION, direction

    def _processline_charmetrics(self, line, charno):
        if line.rstrip() == "EndCharMetrics":
            if self.expected_entries is not None and charno != self.expected_entries:
                # This seems to be a rather common error in AFM files, so we do not raise
                # an exception here, but just graticiously accept the file
                pass
                # raise AFMError("Fewer character metrics than expected")
            return _READ_MAIN, None

        has_name = False
        char = None
        for s in line.split(";"):
            s = s.strip()
            if not s:
               continue
            key, args = s.split(None, 1)
            if key == "C":
                if char is not None:
                    raise AFMError("Cannot define char code twice")
                char = AFMcharmetrics(_parseint(args))
            elif key == "CH":
                if char is not None:
                    raise AFMError("Cannot define char code twice")
                char = AFMcharmetrics(_parsehex(args))
            elif key == "WX" or key == "W0X":
                char.widths[0] = _parsefloat(args), 0
            elif key == "W1X":
                char.widths[1] = _parsefloat(args), 0
            elif key == "WY" or key == "W0Y":
                char.widths[0] = 0, _parsefloat(args)
            elif key == "W1Y":
                char.widths[1] = 0, _parsefloat(args)
            elif key == "W" or key == "W0":
                char.widths[0] = _parsefloats(args, 2)
            elif key == "W1":
                char.widths[1] = _parsefloats(args, 2)
            elif key == "VV":
                char.vvector = _parsefloats(args, 2)
            elif key == "N":
                # XXX: we should check that name is valid (no whitespace, etc.)
                has_name = True
                char.name = _parsestr(args)
            elif key == "B":
                char.bbox = _parsefloats(args, 4)
            elif key == "L":
                successor, ligature = args.split(None, 1)
                char.ligatures.append((_parsestr(successor), ligature))
            else:
                raise AFMError("Undefined command in character widths specification: '%s'", s)
        if char is None:
            raise AFMError("Character metrics not defined")

        self.charmetrics.append(char)
        if has_name:
            self.charmetricsdict[char.name] = char
        return _READ_CHARMETRICS, charno+1

    def _processline_kerndata(self, line):
        try:
            key, args = line.split(None, 1)
        except ValueError:
            key = line
            args = None
        if key == "Comment":
            return _READ_KERNDATA, None
        if key == "StartTrackKern":
            if self.trackkerns is not None:
                raise AFMError("Multiple track kernings data sections")
            self.trackkerns = [None] * _parseint(args)
            return _READ_TRACKKERN, 0
        elif key == "StartKernPairs" or key == "StartKernPairs0":
            if self.kernpairs[0] is not None:
                raise AFMError("Multiple kerning pairs data sections for direction 0")
            if args is not None:
                self.expected_entries = _parseint(args)
            else:
                self.expected_entries = None
            self.kernpairs[0] = []
            return _READ_KERNPAIRS, (0, 0)
        elif key == "StartKernPairs1":
            if self.kernpairs[1] is not None:
                raise AFMError("Multiple kerning pairs data sections for direction 1")
            self.expected_entries = _parseint(args)
            self.kernpairs[1] = []
            return _READ_KERNPAIRS, (1, 0)
        elif key == "EndKernData":
            return _READ_MAIN, None
        else:
            raise AFMError("Unsupported key %s in kerning data section" % key)

    def _processline_trackkern(self, line, i):
        try:
            key, args = line.split(None, 1)
        except ValueError:
            key = line
        if key == "Comment":
            return _READ_TRACKKERN, i
        elif key == "TrackKern":
            if i >= len(self.trackkerns):
                raise AFMError("More track kerning data sets than expected")
            degrees, args = args.split(None, 1)
            self.trackkerns[i] = AFMtrackkern(int(degrees), *_parsefloats(args, 4))
            return _READ_TRACKKERN, i+1
        elif key == "EndTrackKern":
            if i < len(self.trackkerns):
                raise AFMError("Fewer track kerning data sets than expected")
            return _READ_KERNDATA, None
        else:
            raise AFMError("Unsupported key %s in kerning data section" % key)

    def _processline_kernpairs(self, line, xxx_todo_changeme):
        (direction, i) = xxx_todo_changeme
        try:
            key, args = line.split(None, 1)
        except ValueError:
            key = line
        if key == "Comment":
            return _READ_KERNPAIRS, (direction, i)
        elif key == "EndKernPairs":
            if i != self.expected_entries:
                # This seems to be a rather common error in AFM files, so we do not raise
                # an exception here, but just graticiously accept the file
                pass
                # raise AFMError("Fewer kerning pairs than expected")
            return _READ_KERNDATA, None
        else:
            if key == "KP":
                try:
                    name1, name2, x, y = args.split()
                except Exception:
                    raise AFMError("Expecting name1, name2, x, y, got '%s'" % args)
                x = _parsefloat(x)
                y = _parsefloat(y)
            elif key == "KPH":
                try:
                    hex1, hex2, x, y = args.split()
                except Exception:
                    raise AFMError("Expecting <hex1>, <hex2>, x, y, got '%s'" % args)
                name1 = _parsehex(hex1)
                name2 = _parsehex(hex2)
                x = _parsefloat(x)
                y = _parsefloat(y)
            elif key == "KPX":
                try:
                    name1, name2, x = args.split()
                except Exception:
                    raise AFMError("Expecting name1, name2, x, got '%s'" % args)
                x = _parsefloat(x)
                y = 0
            elif key == "KPY":
                try:
                    name1, name2, y = args.split()
                except Exception:
                    raise AFMError("Expecting name1, name2, y, got '%s'" % args)
                x = 0
                y = _parsefloat(y)
            else:
                raise AFMError("Unknown key '%s' in kern pair section" % key)
            kernpair = AFMkernpair(name1, name2, x, y)
            self.kernpairs[direction].append(kernpair)
            if direction:
                self.kernpairsdict1[name1, name2] = kernpair
            else:
                self.kernpairsdict[name1, name2] = kernpair
            return _READ_KERNPAIRS, (direction, i+1)

    def _processline_composites(self, line, i):
        if line == "EndComposites":
            if self.expected_entries is not None and i != self.expected_entries:
                raise AFMError("Fewer composites than expected")
            return _READ_MAIN, None

        name = None
        no = None
        parts = []
        for s in line.split(";"):
            s = s.strip()
            if not s:
               continue
            key, args = s.split(None, 1)
            if key == "CC":
                try:
                    name, no = args.split()
                except Exception:
                    raise AFMError("Expecting name number, got '%s'" % args)
                no = _parseint(no)
            elif key == "PCC":
                try:
                    name1, x, y = args.split()
                except Exception:
                    raise AFMError("Expecting name x y, got '%s'" % args)
                parts.append((name1, _parsefloat(x), _parsefloat(y)))
            else:
                raise AFMError("Unknown key '%s' in composite character data section" % key)
        if len(parts) != no:
            raise AFMError("Wrong number of composite characters")
        return _READ_COMPOSITES, i+1

    def parse(self, f):
         # state of the reader, consisting of 
         #  - the main state, i.e. the type of the section
         #  - a parameter sstate
         state = _READ_START, None
         # Note that we do a line by line processing here, since one
         # of the states (_READ_DIRECTION) can be entered implicitly, i.e.
         # without a corresponding StartDirection section and we thus
         # may need to reprocess a line in the context of the new state
         for line in f:
            line = line[:-1].strip()
            mstate, sstate = state
            if mstate == _READ_START:
                state = self._processline_start(line)
            else: 
                # except for the first line, any empty will be ignored
                if not line:
                   continue
                if mstate == _READ_MAIN:
                    state = self._processline_main(line)
                elif mstate == _READ_DIRECTION:
                    state = self._processline_direction(line, sstate)
                elif mstate == _READ_CHARMETRICS:
                    state = self._processline_charmetrics(line, sstate)
                elif mstate == _READ_KERNDATA:
                    state = self._processline_kerndata(line)
                elif mstate == _READ_TRACKKERN:
                    state = self._processline_trackkern(line, sstate)
                elif mstate == _READ_KERNPAIRS:
                    state = self._processline_kernpairs(line, sstate)
                elif mstate == _READ_COMPOSITES:
                    state = self._processline_composites(line, sstate)
                else:
                    raise AFMError("Undefined state in AFM reader")

    def width_ds(self, glyphname):
        return self.charmetricsdict[glyphname].widths[0][0]

    def width_pt(self, glyphnames, size_pt):
        return sum([self.charmetricsdict[glyphname].widths[0][0] for glyphname in glyphnames])*size_pt/1000.0

    def height_pt(self, glyphnames, size_pt):
        return max([self.charmetricsdict[glyphname].bbox[3] for glyphname in glyphnames])*size_pt/1000.0

    def depth_pt(self, glyphnames, size_pt):
        return min([self.charmetricsdict[glyphname].bbox[1] for glyphname in glyphnames])*size_pt/1000.0

    def resolveligatures(self, glyphnames):
        i = 1
        while i < len(glyphnames):
            for glyphname, replacement in self.charmetricsdict[glyphnames[i-1]].ligatures:
                if glyphname == glyphnames[i]:
                    glyphnames[i-1] = replacement
                    del glyphnames[i]
                    break
            else:
                i += 1
        return glyphnames

    def resolvekernings(self, glyphnames, size_pt=None):
        result = [None]*(2*len(glyphnames)-1)
        for i, glyphname in enumerate(glyphnames):
            result[2*i] = glyphname
            if i:
                kernpair = self.kernpairsdict.get((glyphnames[i-1], glyphname))
                if kernpair:
                    if size_pt is not None:
                        result[2*i-1] = kernpair.x*size_pt/1000.0
                    else:
                        result[2*i-1] = kernpair.x
        return result

    def writePDFfontinfo(self, file, seriffont=False, symbolfont=True):
        flags = 0
        if self.isfixedpitchs[0]:
            flags += 1<<0
        if seriffont:
            flags += 1<<1
        if symbolfont:
            flags += 1<<2
        else:
            flags += 1<<5
        if self.italicangles[0]:
            flags += 1<<6
        file.write("/Flags %d\n" % flags)
        if self.italicangles[0] is not None:
            file.write("/ItalicAngle %d\n" % self.italicangles[0])
        if self.ascender is not None:
            ascent = self.ascender
        elif self.fontbbox is not None:
            ascent = self.fontbbox[3]
        else:
            ascent = 1000 # guessed default
        file.write("/Ascent %d\n" % ascent)
        if self.descender is not None:
            descent = self.descender
        elif self.fontbbox is not None:
            descent = self.fontbbox[3]
        else:
            descent = -200 # guessed default
        file.write("/Descent %d\n" % descent)
        if self.fontbbox is not None:
            file.write("/FontBBox [%d %d %d %d]\n" % tuple(self.fontbbox))
        else:
            # the fontbbox is required, so we have to have to provide some default
            file.write("/FontBBox [0 %d 1000 %d]\n" % (descent, ascent))
        if self.capheight is not None:
            file.write("/CapHeight %d\n" % self.capheight)
        else:
            # the CapHeight is required, so we have to have to provide some default
            file.write("/CapHeight %d\n" % ascent)
        if self.stdvw is not None:
            stemv = self.stdvw
        elif self.weight is not None and ("bold" in self.weight.lower() or "black" in self.weight.lower()):
            stemv = 120 # guessed default
        else:
            stemv = 70 # guessed default
        file.write("/StemV %d\n" % stemv)



from math import sin, cos, radians
from pyx import *

angle = 10
factor = 1.0 / (cos(radians(angle)) + sin(radians(angle)))

cc = canvas.canvas()
cc.stroke(path.rect(-2, -2, 4, 4))

c = canvas.canvas()
for i in range(10):
    c.insert(cc, [trafo.rotate(i*angle), trafo.scale(factor**i)])
c.writeEPSfile("insert")
c.writePDFfile("insert")
c.writeSVGfile("insert")



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2007-2011 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2007-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


import io, struct


class reader:

    def __init__(self, filename):
        self.file = open(filename, "rb")

    def tell(self):
        return self.file.tell()

    def eof(self):
        return self.file.eof()

    def read(self, bytes):
        return self.file.read(bytes)

    def readint(self, bytes=4, signed=0):
        first = 1
        result = 0
        while bytes:
            value = ord(self.file.read(1))
            if first and signed and value > 127:
                value -= 256
            first = 0
            result = 256 * result + value
            bytes -= 1
        return result

    def readint32(self):
        return struct.unpack(">l", self.file.read(4))[0]

    def readuint32(self):
        return struct.unpack(">L", self.file.read(4))[0]

    def readint24(self):
        return struct.unpack(">l", b"\0"+self.file.read(3))[0]

    def readuint24(self):
        return struct.unpack(">L", b"\0"+self.file.read(3))[0]

    def readint16(self):
        return struct.unpack(">h", self.file.read(2))[0]

    def readuint16(self):
        return struct.unpack(">H", self.file.read(2))[0]

    def readchar(self):
        return struct.unpack("b", self.file.read(1))[0]

    def readuchar(self):
        return struct.unpack("B", self.file.read(1))[0]

    def readstring(self, bytes):
        l = self.readuchar()
        assert l <= bytes-1, "inconsistency in file: string too long"
        return self.file.read(bytes-1)[:l]

    def close(self):
        self.file.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return self.file.__exit__(exc_type, exc_value, traceback)


class bytesreader(reader):

    def __init__(self, b):
        self.file = io.BytesIO(b)


class PStokenizer:
    """cursor to read a string token by token"""

    def __init__(self, data, startstring=None, eattokensep=1,
                 tokenseps=" \t\r\n", tokenstarts="()<>[]{}/%",
                 commentchar="%", newlinechars="\r\n"):
        """creates a cursor for the string data

        startstring is a string at which the cursor should start at. The first
        ocurance of startstring is used. When startstring is not in data, an
        exception is raised, otherwise the cursor is set to the position right
        after the startstring. When eattokenseps is set, startstring must be
        followed by a tokensep and this first tokensep is also consumed.
        tokenseps is a string containing characters to be used as token
        separators. tokenstarts is a string containing characters which 
        directly (even without intermediate token separator) start a new token.
        """
        self.data = data
        if startstring is not None:
            self.pos = self.data.index(startstring) + len(startstring)
        else:
            self.pos = 0
        self.tokenseps = tokenseps
        self.tokenstarts = tokenstarts
        self.commentchar = commentchar
        self.newlinechars = newlinechars
        if eattokensep:
            if self.data[self.pos] not in self.tokenstarts:
                if self.data[self.pos] not in self.tokenseps:
                    raise ValueError("cursor initialization string is not followed by a token separator")
                self.pos += 1

    def gettoken(self):
        """get the next token

        Leading token separators and comments are silently consumed. The first token
        separator after the token is also silently consumed."""
        while self.data[self.pos] in self.tokenseps:
            self.pos += 1
        # ignore comments including subsequent whitespace characters
        while self.data[self.pos] == self.commentchar:
            while self.data[self.pos] not in self.newlinechars:
                self.pos += 1
            while self.data[self.pos] in self.tokenseps:
                self.pos += 1
        startpos = self.pos
        while self.data[self.pos] not in self.tokenseps:
            # any character in self.tokenstarts ends the token
            if self.pos>startpos and self.data[self.pos] in self.tokenstarts:
                break
            self.pos += 1
        result = self.data[startpos:self.pos]
        if self.data[self.pos] in self.tokenseps:
            self.pos += 1 # consume a single tokensep
        return result

    def getint(self):
        """get the next token as an integer"""
        return int(self.gettoken())

    def getbytes(self, count):
        """get the next count bytes"""
        startpos = self.pos
        self.pos += count
        return self.data[startpos: self.pos]



class PSbytes_tokenizer(PStokenizer):

    def __init__(self, data, startstring=None, eattokensep=1,
                 tokenseps=b" \t\r\n", tokenstarts=b"()<>[]{}/%",
                 commentchar=b"%", newlinechars=b"\r\n"):
        super().__init__(data, startstring=startstring, eattokensep=eattokensep,
                         tokenseps=tokenseps, tokenstarts=tokenstarts,
                         commentchar=commentchar, newlinechars=newlinechars)



from pyx import *

c = graph.axis.pathaxis(path.curve(0, 0, 3, 0, 1, 4, 4, 4),
                        graph.axis.linear(min=0, max=10))
c.writeEPSfile("minimal")
c.writePDFfile("minimal")
c.writeSVGfile("minimal")



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2002-2011 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2002-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import math
from . import attr, baseclasses, unit

# some helper routines

def _rmatrix(angle):
    phi = math.pi*angle/180.0

    return  ((math.cos(phi), -math.sin(phi)), 
             (math.sin(phi),  math.cos(phi)))

def _rvector(angle, x, y):
    phi = math.pi*angle/180.0

    return  ((1-math.cos(phi))*x + math.sin(phi)    *y,
              -math.sin(phi)   *x + (1-math.cos(phi))*y)


def _mmatrix(angle):
    phi = math.pi*angle/180.0

    return ( (math.cos(phi)*math.cos(phi)-math.sin(phi)*math.sin(phi),
              -2*math.sin(phi)*math.cos(phi)                ),
             (-2*math.sin(phi)*math.cos(phi),
              math.sin(phi)*math.sin(phi)-math.cos(phi)*math.cos(phi) ) )


# trafo: affine transformations

class trafo_pt(baseclasses.deformer):

    """affine transformation (coordinates in constructor in pts)

    Note that though the coordinates in the constructor are in
    pts (which is useful for internal purposes), all other
    methods only accept units in the standard user notation.

    """

    def __init__(self, matrix=((1, 0), (0, 1)), vector=(0, 0)):
        """Return trafo with given transformation matrix and vector.
        """
        self.matrix = matrix
        self.vector = vector

    def __mul__(self, other):
        if isinstance(other, trafo_pt):
            matrix = ( ( self.matrix[0][0]*other.matrix[0][0] +
                         self.matrix[0][1]*other.matrix[1][0],
                         self.matrix[0][0]*other.matrix[0][1] +
                         self.matrix[0][1]*other.matrix[1][1] ),
                       ( self.matrix[1][0]*other.matrix[0][0] +
                         self.matrix[1][1]*other.matrix[1][0],
                         self.matrix[1][0]*other.matrix[0][1] +
                         self.matrix[1][1]*other.matrix[1][1] )
                     )

            vector = ( self.matrix[0][0]*other.vector[0] +
                       self.matrix[0][1]*other.vector[1] +
                       self.vector[0],
                       self.matrix[1][0]*other.vector[0] +
                       self.matrix[1][1]*other.vector[1] +
                       self.vector[1] )

            return trafo_pt(matrix=matrix, vector=vector)
        else:
            raise NotImplementedError("can only multiply two transformations")

    def __str__(self):
        return "[%f %f %f %f %f %f]" % \
               ( self.matrix[0][0], self.matrix[1][0],
                 self.matrix[0][1], self.matrix[1][1],
                 self.vector[0], self.vector[1] )

    def processPS(self, file, writer, context, registry):
        file.write("[%f %f %f %f %f %f] concat\n" % \
                    ( self.matrix[0][0], self.matrix[1][0],
                      self.matrix[0][1], self.matrix[1][1],
                      self.vector[0], self.vector[1] ) )

    def processPDF(self, file, writer, context, registry):
        file.write("%f %f %f %f %f %f cm\n" % \
                    ( self.matrix[0][0], self.matrix[1][0],
                      self.matrix[0][1], self.matrix[1][1],
                      self.vector[0], self.vector[1] ) )

    def processSVGattrs(self, attrs, writer, context, registry):
        assert "transform" not in attrs
        attrs["transform"] = "matrix(%f,%f,%f,%f,%f,%f)" % \
                    ( self.matrix[0][0], -self.matrix[1][0],
                      -self.matrix[0][1], self.matrix[1][1],
                      self.vector[0], -self.vector[1] )

    def apply_pt(self, x_pt, y_pt):
        """apply transformation to point (x_pt, y_pt) in pts"""
        return ( self.matrix[0][0]*x_pt + self.matrix[0][1]*y_pt + self.vector[0],
                 self.matrix[1][0]*x_pt + self.matrix[1][1]*y_pt + self.vector[1] )

    def apply(self, x, y):
        # for the transformation we have to convert to points
        tx, ty = self.apply_pt(unit.topt(x), unit.topt(y))
        return tx * unit.t_pt, ty * unit.t_pt

    def deform(self, path):
        return path.transformed(self)

    def inverse(self):
        det = 1.0*(self.matrix[0][0]*self.matrix[1][1] - self.matrix[0][1]*self.matrix[1][0])
        matrix = ( ( self.matrix[1][1]/det, -self.matrix[0][1]/det),
                   (-self.matrix[1][0]/det,  self.matrix[0][0]/det) )
        return ( trafo_pt(matrix=matrix) *
                 trafo_pt(vector=(-self.vector[0], -self.vector[1])) )

    def mirrored(self, angle):
        return mirror(angle) * self

    def rotated_pt(self, angle, x=None, y=None):
        return rotate_pt(angle, x, y) * self

    def rotated(self, angle, x=None, y=None):
        return rotate(angle, x, y) * self

    def scaled_pt(self, sx, sy=None, x=None, y=None):
        return scale_pt(sx, sy, x, y) * self

    def scaled(self, sx, sy=None, x=None, y=None):
        return scale(sx, sy, x, y) * self

    def slanted_pt(self, a, angle=0, x=None, y=None):
        return slant_pt(a, angle, x, y) * self

    def slanted(self, a, angle=0, x=None, y=None):
        return slant(a, angle, x, y) * self

    def translated_pt(self, x, y):
        return translate_pt(x, y) * self

    def translated(self, x, y):
        return translate(x, y) * self


class trafo(trafo_pt):

    """affine transformation"""

    def __init__(self, matrix=((1,0), (0,1)), vector=(0, 0)):
        trafo_pt.__init__(self,
                          matrix, (unit.topt(vector[0]), unit.topt(vector[1])))

#
# some standard transformations 
#

identity = trafo()

class mirror(trafo):
    def __init__(self, angle=0):
        trafo.__init__(self, matrix=_mmatrix(angle))


class rotate_pt(trafo_pt):
    def __init__(self, angle, x=None, y=None):
        vector = 0, 0
        if x is not None or y is not None:
            if x is None or y is None:
                raise TrafoException("either specify both x and y or none of them")
            vector=_rvector(angle, x, y)

        trafo_pt.__init__(self, matrix=_rmatrix(angle), vector=vector)


class rotate(trafo_pt):
    def __init__(self, angle, x=None, y=None):
        vector = 0, 0 
        if x is not None or y is not None:
            if x is None or y is None:
                raise TrafoException("either specify both x and y or none of them")
            vector=_rvector(angle, unit.topt(x), unit.topt(y))

        trafo_pt.__init__(self, matrix=_rmatrix(angle), vector=vector)


class scale_pt(trafo_pt):
    def __init__(self, sx, sy=None, x=None, y=None):
        if sy is None:
            sy = sx
        vector = 0, 0
        if x is not None or y is not None:
            if x is None or y is None:
                raise TrafoException("either specify both x and y or none of them")
            vector = (1-sx)*x, (1-sy)*y
        trafo_pt.__init__(self, matrix=((sx, 0), (0, sy)), vector=vector)


class scale(trafo):
    def __init__(self, sx, sy=None, x=None, y=None):
        if sy is None:
            sy = sx
        vector = 0, 0
        if x is not None or y is not None:
            if x is None or y is None:
                raise TrafoException("either specify both x and y or none of them")
            vector = (1-sx)*x, (1-sy)*y
        trafo.__init__(self, matrix=((sx, 0), (0, sy)), vector=vector)


class slant_pt(trafo_pt):
    def __init__(self, a, angle=0, x=None, y=None):
        t = ( rotate_pt(-angle, x, y) *
              trafo(matrix=((1, a), (0, 1))) *
              rotate_pt(angle, x, y) )
        trafo_pt.__init__(self, t.matrix, t.vector)


class slant(trafo):
    def __init__(self, a, angle=0, x=None, y=None):
        t = ( rotate(-angle, x, y) *
              trafo(matrix=((1, a), (0, 1))) *
              rotate(angle, x, y) )
        trafo.__init__(self, t.matrix, t.vector)


class translate_pt(trafo_pt):
    def __init__(self, x, y):
        trafo_pt.__init__(self, vector=(x, y))


class translate(trafo):
    def __init__(self, x, y):
        trafo.__init__(self, vector=(x, y))



from pyx import *

g = graph.graphxy(width=8,
                  x=graph.axis.linear(min=0, max=2),
                  y=graph.axis.linear(min=0, max=2),
                  key=graph.key.key(pos="br", dist=0.1))
g.plot([graph.data.function("x(y)=y**4", title=r"$x = y^4$"),
        graph.data.function("x(y)=y**2", title=r"$x = y^2$"),
        graph.data.function("x(y)=y", title=r"$x = y$"),
        graph.data.function("y(x)=x**2", title=r"$y = x^2$"),
        graph.data.function("y(x)=x**4", title=r"$y = x^4$")],
       [graph.style.line([color.gradient.Rainbow])])
g.writeEPSfile("change")
g.writePDFfile("change")
g.writeSVGfile("change")



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2002 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2002 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import os

styles = (("10pt", "article", "10pt", "" , ),
          ("11pt", "article", "11pt", "" , ),
          ("12pt", "article", "12pt", "" , ),
          ("10ptex", "article", "10pt", "\\usepackage{exscale}" , ),
          ("11ptex", "article", "11pt", "\\usepackage{exscale}" , ),
          ("12ptex", "article", "12pt", "\\usepackage{exscale}" , ),
          ("foils17pt", "foils", "17pt", "" , ),
          ("foils20pt", "foils", "20pt", "" , ),
          ("foils25pt", "foils", "25pt", "" , ),
          ("foils30pt", "foils", "30pt", "" , ), )

for style in styles:
    os.system("echo \'%s\n%s\n%s\n%s\'|latex createlfs.tex" % style)



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2006-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


# Just a quick'n'dirty ascii art (I'll do a nice PyX plot later on):
#
#
#      node1 *
#            | \
#            |   \  neighbor2
#            |     \
#            |       \
#  neighbor3 |element * node3
#            |       /
#            |     /
#            |   /  neighbor1
#            | /
#      node2 *


import struct, binascii, zlib, os, tempfile
from . import bbox, baseclasses, color, pdfwriter, unit


class node_pt:

    def __init__(self, coords_pt, value):
        self.coords_pt = coords_pt
        self.value = value


class node(node_pt):

    def __init__(self, coords, value):
        node_pt.__init__(self, [unit.topt(coord) for coord in coords], value)


class element:

    def __init__(self, nodes, neighbors=None):
        self.nodes = nodes
        self.neighbors = neighbors


def coords24bit_pt(coords_pt, min_pt, max_pt):
    return struct.pack(">I", int((coords_pt-min_pt)*16777215.0/(max_pt-min_pt)))[1:]


class PDFGenericResource(pdfwriter.PDFobject):

    def __init__(self, type, name, content):
        pdfwriter.PDFobject.__init__(self, type, name)
        self.content = content

    def write(self, file, writer, registry):
        file.write_bytes(self.content)


class mesh(baseclasses.canvasitem):

    def __init__(self, elements, check=1):
        self.elements = elements
        if check:
            colorspacestring = ""
            for element in elements:
                if len(element.nodes) != 3:
                    raise ValueError("triangular mesh expected")
                try:
                    for node in element.nodes:
                        if not colorspacestring:
                            colorspacestring = node.value.colorspacestring()
                        elif node.value.colorspacestring() != colorspacestring:
                            raise ValueError("color space mismatch")
                except AttributeError:
                    raise ValueError("gray, rgb or cmyk color values expected")
                for node in element.nodes:
                    if len(node.coords_pt) != 2:
                        raise ValueError("two dimensional coordinates expected")

    def bbox(self):
        return bbox.bbox_pt(min([node.coords_pt[0] for element in self.elements for node in element.nodes]),
                            min([node.coords_pt[1] for element in self.elements for node in element.nodes]),
                            max([node.coords_pt[0] for element in self.elements for node in element.nodes]),
                            max([node.coords_pt[1] for element in self.elements for node in element.nodes]))

    def data(self, bbox):
        return b"".join([b"\000" + coords24bit_pt(node.coords_pt[0], bbox.llx_pt, bbox.urx_pt) +
                                   coords24bit_pt(node.coords_pt[1], bbox.lly_pt, bbox.ury_pt) +
                                   node.value.to8bitbytes()
                         for element in self.elements for node in element.nodes])

    def processPS(self, file, writer, context, registry, bbox):
        if writer.meshasbitmap:
            from pyx import bitmap, canvas
            from PIL import Image
            c = canvas.canvas()
            c.insert(self)
            i = Image.open(c.pipeGS("pngalpha", resolution=writer.meshasbitmapresolution))
            i.load()
            b = bitmap.bitmap_pt(self.bbox().llx_pt, self.bbox().lly_pt, i)
            # we slightly shift the bitmap to re-center it, as the bitmap might contain some additional border
            # unfortunately we need to construct another bitmap instance for that ...
            b = bitmap.bitmap_pt(self.bbox().llx_pt + 0.5*(self.bbox().width_pt()-b.bbox().width_pt()),
                                 self.bbox().lly_pt + 0.5*(self.bbox().height_pt()-b.bbox().height_pt()), i)
            b.processPS(file, writer, context, registry, bbox)
        else:
            thisbbox = self.bbox()
            bbox += thisbbox
            file.write("""<< /ShadingType 4
/ColorSpace %s
/BitsPerCoordinate 24
/BitsPerComponent 8
/BitsPerFlag 8
/Decode [%f %f %f %f %s]
/DataSource currentfile /ASCIIHexDecode filter /FlateDecode filter
>> shfill\n""" % (self.elements[0].nodes[0].value.colorspacestring(),
                  thisbbox.llx_pt, thisbbox.urx_pt, thisbbox.lly_pt, thisbbox.ury_pt,
                  " ".join(["0 1" for value in self.elements[0].nodes[0].value.to8bitbytes()])))
            file.write_bytes(binascii.b2a_hex(zlib.compress(self.data(thisbbox))))
            file.write(">\n")

    def processPDF(self, file, writer, context, registry, bbox):
        if writer.meshasbitmap:
            from pyx import bitmap, canvas
            from PIL import Image
            c = canvas.canvas()
            c.insert(self)
            i = Image.open(c.pipeGS("pngalpha", resolution=writer.meshasbitmapresolution))
            i.load()
            b = bitmap.bitmap_pt(self.bbox().llx_pt, self.bbox().lly_pt, i)
            # we slightly shift the bitmap to re-center it, as the bitmap might contain some additional border
            # unfortunately we need to construct another bitmap instance for that ...
            b = bitmap.bitmap_pt(self.bbox().llx_pt + 0.5*(self.bbox().width_pt()-b.bbox().width_pt()),
                                 self.bbox().lly_pt + 0.5*(self.bbox().height_pt()-b.bbox().height_pt()), i)
            b.processPDF(file, writer, context, registry, bbox)
        else:
            thisbbox = self.bbox()
            bbox += thisbbox
            d = self.data(thisbbox)
            if writer.compress:
                filter = "/Filter /FlateDecode\n"
                d = zlib.compress(d)
            else:
                filter = ""
            name = "shading-%s" % id(self)
            shading = PDFGenericResource("shading", name, ("""<<
/ShadingType 4
/ColorSpace %s
/BitsPerCoordinate 24
/BitsPerComponent 8
/BitsPerFlag 8
/Decode [%f %f %f %f %s]
/Length %i
%s>>
stream
""" %            (self.elements[0].nodes[0].value.colorspacestring(),
                  thisbbox.llx_pt, thisbbox.urx_pt, thisbbox.lly_pt, thisbbox.ury_pt,
                  " ".join(["0 1" for value in self.elements[0].nodes[0].value.to8bitbytes()]),
                  len(d), filter)).encode('ascii') + d + b"\nendstream\n")
            registry.add(shading)
            registry.addresource("Shading", name, shading)
            file.write("/%s sh\n" % name)

    def processSVG(self, xml, writer, context, registry, bbox):
        from pyx import bitmap, canvas
        from PIL import Image
        c = canvas.canvas()
        c.insert(self)
        i = Image.open(c.pipeGS("pngalpha", resolution=writer.meshasbitmapresolution))
        i.load()
        b = bitmap.bitmap_pt(self.bbox().llx_pt, self.bbox().lly_pt, i)
        # we slightly shift the bitmap to re-center it, as the bitmap might contain some additional border
        # unfortunately we need to construct another bitmap instance for that ...
        b = bitmap.bitmap_pt(self.bbox().llx_pt + 0.5*(self.bbox().width_pt()-b.bbox().width_pt()),
                             self.bbox().lly_pt + 0.5*(self.bbox().height_pt()-b.bbox().height_pt()), i)
        b.processSVG(xml, writer, context, registry, bbox)



# -*- coding: ISO-8859-1 -*-
#
# Copyright (C) 2011 Michael Schindler <m-schindler@users.sourceforge.net>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from math import atan2, radians
from pyx import unit, attr, normpath
from pyx import path as pathmodule

from .mp_path import mp_endpoint, mp_explicit, mp_given, mp_curl, mp_open, mp_end_cycle, mp_make_choices

# global epsilon (default precision length of metapost, in pt)
_epsilon = 1e-5

def set(epsilon=None):
    global _epsilon
    if epsilon is not None:
        _epsilon = epsilon

################################################################################
# Path knots
################################################################################

class _knot:

    """Internal knot as used in MetaPost (mp.c)"""

    def __init__(self, x_pt, y_pt, ltype, lx_pt, ly_pt, rtype, rx_pt, ry_pt):
        self.x_pt = x_pt
        self.y_pt = y_pt
        self.ltype = ltype
        self.lx_pt = lx_pt
        self.ly_pt = ly_pt
        self.rtype = rtype
        self.rx_pt = rx_pt
        self.ry_pt = ry_pt
        # this is a linked list:
        self.next = self

    def set_left_tension(self, tens):
        self.ly_pt = tens
    def set_right_tension(self, tens):
        self.ry_pt = tens
    def set_left_curl(self, curl):
        self.lx_pt = curl
    def set_right_curl(self, curl):
        self.rx_pt = curl
    set_left_given = set_left_curl
    set_right_given = set_right_curl

    def left_tension(self):
        return self.ly_pt
    def right_tension(self):
        return self.ry_pt
    def left_curl(self):
        return self.lx_pt
    def right_curl(self):
        return self.rx_pt
    left_given = left_curl
    right_given = right_curl

    def linked_len(self):
        """returns the length of a circularly linked list of knots"""
        n = 1
        p = self.next
        while not p is self:
            n += 1
            p = p.next
        return n

    def __repr__(self):
        result = ""
        # left
        if self.ltype == mp_endpoint:
            pass
        elif self.ltype == mp_explicit:
            result += "{explicit %s %s}" % (self.lx_pt, self.ly_pt)
        elif self.ltype == mp_given:
            result += "{given %g tens %g}" % (self.lx_pt, self.ly_pt)
        elif self.ltype == mp_curl:
            result += "{curl %g tens %g}" % (self.lx_pt, self.ly_pt)
        elif self.ltype == mp_open:
            result += "{open tens %g}" % (self.ly_pt)
        elif self.ltype == mp_end_cycle:
            result += "{cycle tens %g}" % (self.ly_pt)
        result += "(%g %g)" % (self.x_pt, self.y_pt)
        # right
        if self.rtype == mp_endpoint:
            pass
        elif self.rtype == mp_explicit:
            result += "{explicit %g %g}" % (self.rx_pt, self.ry_pt)
        elif self.rtype == mp_given:
            result += "{given %g tens %g}" % (self.rx_pt, self.ry_pt)
        elif self.rtype == mp_curl:
            result += "{curl %g tens %g}" % (self.rx_pt, self.ry_pt)
        elif self.rtype == mp_open:
            result += "{open tens %g}" % (self.ry_pt)
        elif self.rtype == mp_end_cycle:
            result += "{cycle tens %g}" % (self.ry_pt)
        return result

class beginknot_pt(_knot):

    """A knot which interrupts a path, or which allows to continue it with a straight line"""

    def __init__(self, x_pt, y_pt, curl=1, angle=None):
        if angle is None:
            type, value = mp_curl, curl
        else:
            type, value = mp_given, angle
        # tensions are modified by the adjacent curve, but default is 1
        _knot.__init__(self, x_pt, y_pt, mp_endpoint, None, None, type, value, 1)

class beginknot(beginknot_pt):

    def __init__(self, x, y, curl=1, angle=None):
        if not (angle is None):
            angle = radians(angle)
        beginknot_pt.__init__(self, unit.topt(x), unit.topt(y), curl, angle)

startknot = beginknot

class endknot_pt(_knot):

    """A knot which interrupts a path, or which allows to continue it with a straight line"""

    def __init__(self, x_pt, y_pt, curl=1, angle=None):
        if angle is None:
            type, value = mp_curl, curl
        else:
            type, value = mp_given, angle
        # tensions are modified by the adjacent curve, but default is 1
        _knot.__init__(self, x_pt, y_pt, type, value, 1, mp_endpoint, None, None)

class endknot(endknot_pt):

    def __init__(self, x, y, curl=1, angle=None):
        if not (angle is None):
            angle = radians(angle)
        endknot_pt.__init__(self, unit.topt(x), unit.topt(y), curl, angle)

class smoothknot_pt(_knot):

    """A knot with continous tangent and "mock" curvature."""

    def __init__(self, x_pt, y_pt):
        # tensions are modified by the adjacent curve, but default is 1
        _knot.__init__(self, x_pt, y_pt, mp_open, None, 1, mp_open, None, 1)

class smoothknot(smoothknot_pt):

    def __init__(self, x, y):
        smoothknot_pt.__init__(self, unit.topt(x), unit.topt(y))

knot = smoothknot

class roughknot_pt(_knot):

    """A knot with noncontinous tangent."""

    def __init__(self, x_pt, y_pt, lcurl=1, rcurl=None, langle=None, rangle=None):
        """Specify either the relative curvatures, or tangent angles left (l)
        or right (r) of the point."""
        if langle is None:
            ltype, lvalue = mp_curl, lcurl
        else:
            ltype, lvalue = mp_given, langle
        if rcurl is not None:
            rtype, rvalue = mp_curl, rcurl
        elif rangle is not None:
            rtype, rvalue = mp_given, rangle
        else:
            rtype, rvalue = ltype, lvalue
        # tensions are modified by the adjacent curve, but default is 1
        _knot.__init__(self, x_pt, y_pt, ltype, lvalue, 1, rtype, rvalue, 1)

class roughknot(roughknot_pt):

    def __init__(self, x, y, lcurl=1, rcurl=None, langle=None, rangle=None):
        if langle is not None:
            langle = radians(langle)
        if rangle is not None:
            rangle = radians(rangle)
        roughknot_pt.__init__(self, unit.topt(x), unit.topt(y), lcurl, rcurl, langle, rangle)

################################################################################
# Path links
################################################################################

class _link:
    def set_knots(self, left_knot, right_knot):
        """Sets the internal properties of the metapost knots"""
        pass

class line(_link):

    """A straight line"""

    def __init__(self, keepangles=False):
        """The option keepangles will guarantee a continuous tangent. The
        curvature may become discontinuous, however"""
        self.keepangles = keepangles

    def set_knots(self, left_knot, right_knot):
        left_knot.rtype = mp_endpoint
        right_knot.ltype = mp_endpoint
        left_knot.rx_pt, left_knot.ry_pt = None, None
        right_knot.lx_pt, right_knot.ly_pt = None, None
        if self.keepangles:
            angle = atan2(right_knot.y_pt-left_knot.y_pt, right_knot.x_pt-left_knot.x_pt)
            left_knot.ltype = mp_given
            left_knot.set_left_given(angle)
            right_knot.rtype = mp_given
            right_knot.set_right_given(angle)


class controlcurve_pt(_link):

    """A cubic Bezier curve which has its control points explicity set"""

    def __init__(self, lcontrol_pt, rcontrol_pt):
        """The control points at the beginning (l) and the end (r) must be
        coordinate pairs"""
        self.lcontrol_pt = lcontrol_pt
        self.rcontrol_pt = rcontrol_pt

    def set_knots(self, left_knot, right_knot):
        left_knot.rtype = mp_explicit
        right_knot.ltype = mp_explicit
        left_knot.rx_pt, left_knot.ry_pt = self.lcontrol_pt
        right_knot.lx_pt, right_knot.ly_pt = self.rcontrol_pt

class controlcurve(controlcurve_pt):

    def __init__(self, lcontrol, rcontrol):
        controlcurve_pt.__init__(self, (unit.topt(lcontrol[0]), unit.topt(lcontrol[1])),
                                       (unit.topt(rcontrol[0]), unit.topt(rcontrol[1])))


class tensioncurve(_link):

    """A yet unspecified cubic Bezier curve"""

    def __init__(self, ltension=1, latleast=False, rtension=None, ratleast=None):
        """The tension parameters indicate the tensions at the beginning (l)
        and the end (r) of the curve. Set the parameters (l/r)atleast to True
        if you want to avoid inflection points."""
        if rtension is None:
            rtension = ltension
        if ratleast is None:
            ratleast = latleast
        # make sure that tension >= 0.75 (p. 9 mpman.pdf)
        self.ltension = max(0.75, abs(ltension))
        self.rtension = max(0.75, abs(rtension))
        if latleast:
            self.ltension = -self.ltension
        if ratleast:
            self.rtension = -self.rtension

    def set_knots(self, left_knot, right_knot):
        if left_knot.rtype <= mp_explicit or right_knot.ltype <= mp_explicit:
            raise Exception("metapost curve with given tension cannot have explicit knots")
        left_knot.set_right_tension(self.ltension)
        right_knot.set_left_tension(self.rtension)

curve = tensioncurve


################################################################################
# Path creation class
################################################################################

class path(pathmodule.path):

    """A MetaPost-like path, which finds an optimal way through given points.

    At points, you can either specify a given tangent direction (angle in
    degrees) or a certain "curlyness" (relative to the curvature at the other
    end of a curve), or nothing. In the latter case, both the tangent and the
    "mock" curvature (an approximation to the real curvature, introduced by
    J.D. Hobby in MetaPost) will be continuous.

    The shape of the cubic Bezier curves between two points is controlled by
    its "tension", unless you choose to set the control points manually."""

    def __init__(self, elems, epsilon=None):
        """elems should contain metapost knots or links"""
        if epsilon is None:
            epsilon = _epsilon
        knots = []
        is_closed = True
        for i, elem in enumerate(elems):
            if isinstance(elem, _link):
                elem.set_knots(elems[i-1], elems[(i+1)%len(elems)])
            elif isinstance(elem, _knot):
                knots.append(elem)
                if elem.ltype == mp_endpoint or elem.rtype == mp_endpoint:
                    is_closed = False

        # link the knots among each other
        for i in range(len(knots)):
            knots[i-1].next = knots[i]

        # determine the control points
        mp_make_choices(knots[0], epsilon)

        pathmodule.path.__init__(self)
        # build up the path
        do_moveto = True
        do_lineto = False
        do_curveto = False
        prev = None
        for i, elem in enumerate(elems):
            if isinstance(elem, _link):
                do_moveto = False
                if isinstance(elem, line):
                    do_lineto, do_curveto = True, False
                else:
                    do_lineto, do_curveto = False, True
            elif isinstance(elem, _knot):
                if do_moveto:
                    self.append(pathmodule.moveto_pt(elem.x_pt, elem.y_pt))
                if do_lineto:
                    self.append(pathmodule.lineto_pt(elem.x_pt, elem.y_pt))
                elif do_curveto:
                    self.append(pathmodule.curveto_pt(prev.rx_pt, prev.ry_pt, elem.lx_pt, elem.ly_pt, elem.x_pt, elem.y_pt))
                do_moveto = True
                do_lineto = False
                do_curveto = False
                prev = elem

        # close the path if necessary
        if knots[0].ltype == mp_explicit:
            elem = knots[0]
            if do_lineto and is_closed:
                self.append(pathmodule.closepath())
            elif do_curveto:
                self.append(pathmodule.curveto_pt(prev.rx_pt, prev.ry_pt, elem.lx_pt, elem.ly_pt, elem.x_pt, elem.y_pt))
                if is_closed:
                    self.append(pathmodule.closepath())




# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2007-2011 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2007-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


from pyx import bbox, font, config
from . import tfmfile, vffile

class TeXFontError(Exception): pass

class TeXfont:

    def __init__(self, name, c, q, d, tfmconv, pyxconv, debug=0):
        self.name = name
        self.q = q                  # desired size of font (fix_word) in TeX points
        self.d = d                  # design size of font (fix_word) in TeX points
        self.tfmconv = tfmconv      # conversion factor from tfm units to dvi units
        self.pyxconv = pyxconv      # conversion factor from dvi units to PostScript points
        with config.open(self.name, [config.format.tfm]) as file:
            self.TFMfile = tfmfile.TFMfile(file, debug)

        # We only check for equality of font checksums if none of them
        # is zero. The case c == 0 happend in some VF files and
        # according to the VFtoVP documentation, paragraph 40, a check
        # is only performed if TFMfile.checksum > 0. Anyhow, being
        # more generous here seems to be reasonable
        if self.TFMfile.checksum != c and self.TFMfile.checksum > 0 and c > 0:
            raise TeXFontError("check sums do not agree: %d vs. %d" %
                               (self.TFMfile.checksum, c))

        # Check whether the given design size matches the one defined in the tfm file
        if abs(self.TFMfile.designsize - d) > 4: # XXX: why the deviation?
            raise TeXFontError("design sizes do not agree: %d vs. %d" % (self.TFMfile.designsize, d))
        #if q < 0 or q > 134217728:
        #    raise TeXFontError("font '%s' not loaded: bad scale" % self.name)
        if d < 0 or d > 134217728:
            raise TeXFontError("font '%s' not loaded: bad design size" % self.name)

    def __str__(self):
        return "font %s designed at %g TeX pts used at %g TeX pts" % (self.name, 
                                                                      16.0*self.d/16777216,
                                                                      16.0*self.q/16777216)

    def getsize_pt(self):
        """ return size of font in (PS) points """
        # The factor 16L/16777216L=2**(-20) converts a fix_word (here self.q)
        # to the corresponding float. Furthermore, we have to convert from TeX
        # points to points, hence the factor 72/72.27.
        return 72/72.27 * 16*self.q/16777216

    def _convert_tfm_to_dvi(self, length):
        # doing the integer math with long integers will lead to different roundings
        # return 16*length*int(round(self.q*self.tfmconv))/16777216

        # Knuth instead suggests the following algorithm based on 4 byte integer logic only
        # z = int(round(self.q*self.tfmconv))
        # b0, b1, b2, b3 = [ord(c) for c in struct.pack(">L", length)]
        # assert b0 == 0 or b0 == 255
        # shift = 4
        # while z >= 8388608:
        #     z >>= 1
        #     shift -= 1
        # assert shift >= 0
        # result = ( ( ( ( ( b3 * z ) >> 8 ) + ( b2 * z ) ) >> 8 ) + ( b1 * z ) ) >> shift
        # if b0 == 255:
        #     result = result - (z << (8-shift))

        # however, we can simplify this using a single long integer multiplication,
        # but take into account the transformation of z
        z = int(round(self.q*self.tfmconv))
        assert -16777216 <= length < 16777216 # -(1 << 24) <= length < (1 << 24)
        assert z < 134217728 # 1 << 27
        shift = 20 # 1 << 20
        while z >= 8388608: # 1 << 23
            z >>= 1
            shift -= 1
        # length*z is a long integer, but the result will be a regular integer
        return int(length*int(z) >> shift)

    def _convert_tfm_to_pt(self, length):
        return (16*int(round(length*float(self.q)*self.tfmconv))/16777216) * self.pyxconv

    # routines returning lengths as integers in dvi units

    def getwidth_dvi(self, charcode):
        return self._convert_tfm_to_dvi(self.TFMfile.width[self.TFMfile.char_info[charcode].width_index])

    def getheight_dvi(self, charcode):
        return self._convert_tfm_to_dvi(self.TFMfile.height[self.TFMfile.char_info[charcode].height_index])

    def getdepth_dvi(self, charcode):
        return self._convert_tfm_to_dvi(self.TFMfile.depth[self.TFMfile.char_info[charcode].depth_index])

    def getitalic_dvi(self, charcode):
        return self._convert_tfm_to_dvi(self.TFMfile.italic[self.TFMfile.char_info[charcode].italic_index])

    # routines returning lengths as floats in PostScript points

    def getwidth_pt(self, charcode):
        return self._convert_tfm_to_pt(self.TFMfile.width[self.TFMfile.char_info[charcode].width_index])

    def getheight_pt(self, charcode):
        return self._convert_tfm_to_pt(self.TFMfile.height[self.TFMfile.char_info[charcode].height_index])

    def getdepth_pt(self, charcode):
        return self._convert_tfm_to_pt(self.TFMfile.depth[self.TFMfile.char_info[charcode].depth_index])

    def getitalic_pt(self, charcode):
        return self._convert_tfm_to_pt(self.TFMfile.italic[self.TFMfile.char_info[charcode].italic_index])

    def text_pt(self, x_pt, y_pt, charcodes, fontmap=None):
        return TeXtext_pt(self, x_pt, y_pt, charcodes, self.getsize_pt(), fontmap=fontmap)

    def getMAPline(self, fontmap):
        if self.name not in fontmap:
            raise RuntimeError("missing font information for '%s'; check fontmapping file(s)" % self.name)
        return fontmap[self.name]


class virtualfont(TeXfont):

    def __init__(self, name, file, c, q, d, tfmconv, pyxconv, debug=0):
        TeXfont.__init__(self, name, c, q, d, tfmconv, pyxconv, debug)
        self.vffile = vffile.vffile(file, 1.0*q/d, tfmconv, pyxconv, debug > 1)

    def getfonts(self):
        """ return fonts used in virtual font itself """
        return self.vffile.getfonts()

    def getchar(self, cc):
        """ return dvi chunk corresponding to char code cc """
        return self.vffile.getchar(cc)

    def text_pt(self, *args, **kwargs):
        raise RuntimeError("you don't know what you're doing")


class TeXtext_pt(font.text_pt):

    def __init__(self, font, x_pt, y_pt, charcodes, size_pt, fontmap=None):
        self.font = font
        self.x_pt = x_pt
        self.y_pt = y_pt
        self.charcodes = charcodes
        self.size_pt = size_pt
        self.fontmap = fontmap

        self.width_pt = sum([self.font.getwidth_pt(charcode) for charcode in charcodes])
        self.height_pt = max([self.font.getheight_pt(charcode) for charcode in charcodes])
        self.depth_pt = max([self.font.getdepth_pt(charcode) for charcode in charcodes])

        self._bbox = bbox.bbox_pt(self.x_pt, self.y_pt-self.depth_pt, self.x_pt+self.width_pt, self.y_pt+self.height_pt)

    def bbox(self):
        return self._bbox

    def _text(self, writer):
        if self.fontmap is not None:
            mapline = self.font.getMAPline(self.fontmap)
        else:
            mapline = self.font.getMAPline(writer.getfontmap())
        font = mapline.getfont()
        return font.text_pt(self.x_pt, self.y_pt, self.charcodes, self.size_pt, decoding=mapline.getencoding(), slant=mapline.slant, ignorebbox=True)

    def textpath(self):
        from pyx import pswriter
        return self._text(pswriter._PSwriter()).textpath()

    def processPS(self, file, writer, context, registry, bbox):
        bbox += self.bbox()
        self._text(writer).processPS(file, writer, context, registry, bbox)

    def processPDF(self, file, writer, context, registry, bbox):
        bbox += self.bbox()
        self._text(writer).processPDF(file, writer, context, registry, bbox)

    def processSVG(self, xml, writer, context, registry, bbox):
        bbox += self.bbox()
        self._text(writer).processSVG(xml, writer, context, registry, bbox)




#!/usr/bin/env python
from pyx import *

g = graph.graphxyz(size=4, z=graph.axis.lin(min=0.001))
g.plot(graph.data.file("grid.dat", x=1, y=2, z=3), [graph.style.grid()])
g.writeEPSfile("grid")
g.writePDFfile("grid")
g.writeSVGfile("grid")



from pyx import *

p1 = path.curve(0, 0, 1, 0, 1, 1, 2, 1)
p2 = path.circle(1, 0.5, 0.5)

(a1, a2), (b1, b2) = p1.intersect(p2)

x1, y1 = p1.at(a1)
x2, y2 = p1.at(a2)

c = canvas.canvas()
c.fill(path.circle(x1, y1, 0.1), [color.rgb.blue])
c.fill(path.circle(x2, y2, 0.1), [color.rgb.blue])
c.stroke(p1, [color.rgb.red])
c.stroke(p2, [color.rgb.green])
c.writeEPSfile("intersect")
c.writePDFfile("intersect")
c.writeSVGfile("intersect")



from pyx import *
from pyx.metapost.path import beginknot, endknot, smoothknot, tensioncurve

p1, p2, p3, p4, p5 = (0, 0), (2, 1.33), (1.3, 3), (0.33, 2.33), (1, 1.67)
openpath = metapost.path.path([
    beginknot(*p1), tensioncurve(), smoothknot(*p2), tensioncurve(),
    smoothknot(*p3), tensioncurve(), smoothknot(*p4), tensioncurve(),
    endknot(*p5)])
closedpath = metapost.path.path([
    smoothknot(*p1), tensioncurve(), smoothknot(*p2), tensioncurve(),
    smoothknot(*p3), tensioncurve(), smoothknot(*p4), tensioncurve(),
    smoothknot(*p5), tensioncurve()])
c = canvas.canvas()
for p in [p1, p2, p3, p4, p5]:
    c.fill(path.circle(p[0], p[1], 0.05), [color.rgb.red])
    c.fill(path.circle(p[0], p[1], 0.05), [color.rgb.red, trafo.translate(2, 0)])
c.stroke(openpath)
c.stroke(closedpath, [trafo.translate(2, 0)])

c.writeEPSfile("metapost")
c.writePDFfile("metapost")
c.writeSVGfile("metapost")



#!/usr/bin/env python
from pyx import *

g = graph.graphxyz(size=4, x2=None, y2=None)
g.plot(graph.data.file("surface.dat", x=1, y=2, z=3), [graph.style.surface()])
g.writeEPSfile("surface")
g.writePDFfile("surface")
g.writeSVGfile("surface")
